/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#ifndef __LIS_FORTRAN_H__
#define __LIS_FORTRAN_H__

typedef long			LIS_MATRIX_F;
typedef long			LIS_VECTOR_F;
typedef long			LIS_SOLVER_F;

#ifdef USE_MPI
	#include <mpi.h>
	typedef MPI_Fint	LIS_Comm_f;
#else
	typedef int			LIS_Comm_f;
#endif

#define LIS_V2P(a)	(*(long *)(a))
#define LIS_P2V(a)	((long)(a))

/**************/
/* MATRIX     */
/**************/
#define lis_matrix_create_f F77_FUNC_(lis_matrix_create_f, LIS_MATRIX_CREATE_F) 
#define lis_matrix_duplicate_f F77_FUNC_(lis_matrix_duplicate_f, LIS_MATRIX_DUPLICATE_F) 
#define lis_matrix_destroy_f F77_FUNC_(lis_matrix_destroy_f, LIS_MATRIX_DESTROY_F) 
#define lis_matrix_set_size_f F77_FUNC_(lis_matrix_set_size_f, LIS_MATRIX_SET_SIZE_F) 
#define lis_matrix_get_size_f F77_FUNC_(lis_matrix_get_size_f, LIS_MATRIX_GET_SIZE_F) 
#define lis_matrix_get_range_f F77_FUNC_(lis_matrix_get_range_f, LIS_MATRIX_GET_RANGE_F) 
#define lis_matrix_set_value_f F77_FUNC_(lis_matrix_set_value_f, LIS_MATRIX_SET_VALUE_F) 
#define lis_matrix_get_type_f F77_FUNC_(lis_matrix_get_type_f, LIS_MATRIX_GET_TYPE_F) 
#define lis_matrix_set_type_f F77_FUNC_(lis_matrix_set_type_f, LIS_MATRIX_SET_TYPE_F)
#define lis_matrix_assemble_f F77_FUNC_(lis_matrix_assemble_f, LIS_MATRIX_ASSEMBLE_F) 
/**************/
/* VECTOR     */
/**************/
#define lis_vector_create_f F77_FUNC_(lis_vector_create_f, LIS_VECTOR_CREATE_F) 
#define lis_vector_duplicate_f F77_FUNC_(lis_vector_duplicate_f, LIS_VECTOR_DUPLICATE_F) 
#define lis_vector_destroy_f F77_FUNC_(lis_vector_destroy_f, LIS_VECTOR_DESTROY_F) 
#define lis_vector_set_size_f F77_FUNC_(lis_vector_set_size_f, LIS_VECTOR_SET_SIZE_F) 
#define lis_vector_get_size_f F77_FUNC_(lis_vector_get_size_f, LIS_VECTOR_GET_SIZE_F) 
#define lis_vector_get_range_f F77_FUNC_(lis_vector_get_range_f, LIS_VECTOR_GET_RANGE_F) 
#define lis_vector_set_value_f F77_FUNC_(lis_vector_set_value_f, LIS_VECTOR_SET_VALUE_F) 
#define lis_vector_set_values_f F77_FUNC_(lis_vector_set_values_f, LIS_VECTOR_SET_VALUES_F) 
#define lis_vector_get_value_f F77_FUNC_(lis_vector_get_value_f, LIS_VECTOR_GET_VALUE_F) 
#define lis_vector_get_values_f F77_FUNC_(lis_vector_get_values_f, LIS_VECTOR_GET_VALUES_F) 
#define lis_vector_print_f F77_FUNC_(lis_vector_print_f, LIS_VECTOR_PRINT_F) 
#define lis_vector_set_all_f F77_FUNC_(lis_vector_set_all_f, LIS_VECTOR_SET_ALL_F) 
#define lis_vector_axpy_f F77_FUNC_(lis_vector_axpy_f, LIS_VECTOR_AXPY_F) 
#define lis_vector_xpay_f F77_FUNC_(lis_vector_xpay_f, LIS_VECTOR_XPAY_F) 
#define lis_vector_axpyz_f F77_FUNC_(lis_vector_axpyz_f, LIS_VECTOR_AXPYZ_F) 
#define lis_vector_copy_f F77_FUNC_(lis_vector_copy_f, LIS_VECTOR_COPY_F) 
#define lis_vector_scale_f F77_FUNC_(lis_vector_scale_f, LIS_VECTOR_SCALE_F) 
#define lis_vector_pmul_f F77_FUNC_(lis_vector_pmul_f, LIS_VECTOR_PMUL_F) 
#define lis_vector_pdiv_f F77_FUNC_(lis_vector_pdiv_f, LIS_VECTOR_PDIV_F) 
#define lis_vector_abs_f F77_FUNC_(lis_vector_abs_f, LIS_VECTOR_ABS_F) 
#define lis_vector_reciprocal_f F77_FUNC_(lis_vector_reciprocal_f, LIS_VECTOR_RECIPROCAL_F) 
#define lis_vector_shift_f F77_FUNC_(lis_vector_shift_f, LIS_VECTOR_SHIFT_F) 
#define lis_vector_dot_f F77_FUNC_(lis_vector_dot_f, LIS_VECTOR_DOT_F) 
#define lis_vector_nrm2_f F77_FUNC_(lis_vector_nrm2_f, LIS_VECTOR_NRM2_F) 
#define lis_vector_sum_f F77_FUNC_(lis_vector_sum_f, LIS_VECTOR_SUM_F) 
#define lis_is_vector_null_f F77_FUNC_(lis_is_vector_null_f, LIS_IS_VECTOR_NULL_F) 
/**************/
/* SOLVER     */
/**************/
#define lis_solver_create_f F77_FUNC_(lis_solver_create_f, LIS_SOLVER_CREATE_F) 
#define lis_solver_destroy_f F77_FUNC_(lis_solver_destroy_f, LIS_SOLVER_DESTROY_F) 
#define lis_solve_f F77_FUNC_(lis_solve_f, LIS_SOLVE_F) 
#define lis_solver_set_option_f F77_FUNC_(lis_solver_set_option_f, LIS_SOLVER_SET_OPTION_F) 
#define lis_solver_get_iters_f F77_FUNC_(lis_solver_get_iters_f, LIS_SOLVER_GET_ITERS_F) 
#define lis_solver_get_times_f F77_FUNC_(lis_solver_get_times_f, LIS_SOLVER_GET_TIMES_F) 
#define lis_solver_get_residualnorm_f F77_FUNC_(lis_solver_get_residualnorm_f, LIS_SOLVER_GET_RESIDUALNORM_F) 
#define lis_solver_get_solver_f F77_FUNC_(lis_solver_get_solver_f, LIS_SOLVER_GET_SOLVER_F) 
#define lis_get_solvername_f F77_FUNC_(lis_get_solvername_f, LIS_GET_SOLVERNAME_F) 
#define lis_solver_set_optionC_f F77_FUNC_(lis_solver_set_optionc_f, LIS_SOLVER_SET_OPTIONC_F) 
/**************/
/* MATVEC     */
/**************/
#define lis_matvec_f F77_FUNC_(lis_matvec_f, LIS_MATVEC_F) 
#define lis_matvect_f F77_FUNC_(lis_matvect_f, LIS_MATVECT_F) 
/**************/
/* SYSTEM     */
/**************/
#define lis_finitialize_f F77_FUNC_(lis_finitialize_f, LIS_FINITIALIZE_F) 
#define lis_finalize_f F77_FUNC_(lis_finalize_f, LIS_FINALIZE_F) 
#define lis_wtime_f F77_FUNC_(lis_wtime_f , LIS_WTIME_F)
#define CHKERR_f F77_FUNC_(chkerr_f, CHKERR_F) 
#define lis_set_argv_begin_f F77_FUNC_(lis_set_argv_begin_f, LIS_SET_ARGV_BEGIN_F) 
#define lis_set_argv_f F77_FUNC_(lis_set_argv_f, LIS_SET_ARGV_F) 
#define lis_set_argv_end_f F77_FUNC_(lis_set_argv_end_f, LIS_SET_ARGV_END_F) 
#define lis_arg2args_f F77_FUNC_(lis_arg2args_f, LIS_ARG2ARGS_F) 
#define lis_input_f F77_FUNC_(lis_input_f, LIS_INPUT_F) 
#define lis_output_vector_f F77_FUNC_(lis_output_vector_f, LIS_OUTPUT_VECTOR_F) 
#define lis_output_residual_history_f F77_FUNC_(lis_output_residual_history_f, LIS_OUTPUT_RESIDUAL_HISTORY_F) 

#ifdef __cplusplus
extern "C"
{
#endif

#ifdef __cplusplus
}
#endif

#endif
