/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#ifndef __LIS_PRECISION_H__
#define __LIS_PRECISION_H__

#define SPLITTER 134217729.0

typedef struct
{
	double	hi;
	double	lo;
} LIS_QDOUBLE;

typedef struct
{
	double	*hi;
	double	*lo;
} LIS_DOUBLE_DOUBLE;

#ifdef USE_SSE2
	#define LIS_QUAD_DECLAR			__m128d bh,ch,sh,wh,th,bl,cl,sl,wl,tl,p1,p2,t0,t1,t2,eh
	#define LIS_QUAD_DECLAR_DIV		__m128d xb,xc,xs,xe,xw,xq,xr,xt
#else
	#define LIS_QUAD_DECLAR			double p1,p2,t,bhi,blo,chi,clo,sh,eh,sl,el,th,tl
	#define LIS_QUAD_DECLAR_DIV		double q1,q2,s1,s2,t,p1,p2,chi,clo,bhi,blo; \
									LIS_QDOUBLE r
#endif

/* d = a + b x c */
#define LIS_QUAD_FMA(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				p1 =  (b_hi) * (c_hi); \
				t   = SPLITTER * (b_hi); \
				bhi = t - (t-(b_hi)); \
				blo = (b_hi) - bhi; \
				t   = SPLITTER * (c_hi); \
				chi = t - (t-(c_hi)); \
				clo = (c_hi) - chi; \
				p2 = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo; \
				p2 += ((b_hi) * (c_lo)); \
				p2 += ((b_lo) * (c_hi)); \
				chi = p1 + p2; \
				p2  = p2  - (chi - p1); \
				sh =  (a_hi) + chi; \
				th =  sh - (a_hi); \
				eh =  ((a_hi) - (sh - th)) + (chi - th); \
				sl =  (a_lo) + p2; \
				tl =  sl - (a_lo); \
				el =  ((a_lo) - (sl - tl)) + (p2 - tl); \
				eh += sl; \
				th = sh + eh; \
				eh = eh - (th - sh); \
				eh += el; \
				(d_hi) = th + eh; \
				(d_lo) = eh  - ((d_hi) - th)


#define LIS_QUAD_FMAD(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c) \
				p1  = (b_hi) * (c); \
				t   = SPLITTER * (b_hi); \
				bhi = t - (t-(b_hi)); \
				blo = (b_hi) - bhi; \
				t   = SPLITTER * (c); \
				chi = t - (t-(c)); \
				clo = (c) - chi; \
				p2  = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo; \
				p2 += ((b_lo) * (c)); \
				chi = p1 + p2; \
				p2  = p2  - (chi - p1); \
				sh =  (a_hi) + chi; \
				th =  sh - (a_hi); \
				eh =  ((a_hi) - (sh - th)) + (chi - th); \
				sl =  (a_lo) + p2; \
				tl =  sl - (a_lo); \
				el =  ((a_lo) - (sl - tl)) + (p2 - tl); \
				eh += sl; \
				th = sh + eh; \
				eh = eh - (th - sh); \
				eh += el; \
				(d_hi) = th + eh; \
				(d_lo) = eh  - ((d_hi) - th)


#define LIS_QUAD_FMA_SSE2(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				sh  = _mm_set_pd(SPLITTER,SPLITTER); \
				ch = _mm_set_pd((b_hi),(c_hi)); \
				bh  = _mm_unpackhi_pd(ch,ch); \
				cl = _mm_set_pd((c_lo),(b_lo)); \
				bh  = _mm_mul_sd(bh,ch); \
				sh  = _mm_mul_pd(sh,ch); \
				th  = _mm_sub_pd(sh,ch); \
				sh  = _mm_sub_pd(sh,th); \
				th  = _mm_mul_pd(ch,cl); \
				ch = _mm_sub_pd(ch,sh); \
				eh  = _mm_unpackhi_pd(sh,ch); \
				wh  = _mm_unpacklo_pd(sh,ch); \
				ch = _mm_unpackhi_pd(ch,wh); \
				wh  = _mm_mul_pd(wh,eh); \
				sh  = _mm_mul_pd(sh,ch); \
				ch = _mm_unpackhi_pd(sh,sh); \
				cl = _mm_unpackhi_pd(wh,wh); \
				eh  = _mm_unpackhi_pd(th,th); \
				wh  = _mm_sub_sd(wh,bh); \
				wh  = _mm_add_sd(wh,ch); \
				wh  = _mm_add_sd(wh,sh); \
				wh  = _mm_add_sd(wh,cl); \
				wh  = _mm_add_sd(wh,eh); \
				wh  = _mm_add_sd(wh,th); \
				sh  = _mm_add_sd(wh,bh); \
				th  = _mm_sub_sd(sh,bh); \
				wh  = _mm_sub_sd(wh,th); \
				bl  = _mm_set_pd((a_lo),(a_hi)); \
				sh  = _mm_unpacklo_pd(sh,wh); \
				t0 = _mm_add_pd(bl,sh); \
				eh = _mm_sub_pd(t0,bl); \
				th = _mm_sub_pd(t0,eh); \
				sh = _mm_sub_pd(sh,eh); \
				bl = _mm_sub_pd(bl,th); \
				bl = _mm_add_pd(bl,sh); \
				eh = _mm_unpackhi_pd(bl,bl); \
				sh = _mm_unpackhi_pd(t0,t0); \
				th = t0; \
				bl = _mm_add_sd(bl,sh); \
				th = _mm_add_sd(th,bl); \
				t0 = _mm_sub_sd(th,t0); \
				bl = _mm_sub_sd(bl,t0); \
				bl = _mm_add_sd(bl,eh); \
				sh = _mm_add_sd(th,bl); \
				_mm_store_sd(&(d_hi),sh); \
				sh = _mm_sub_sd(sh,th); \
				bl = _mm_sub_sd(bl,sh); \
				_mm_store_sd(&(d_lo),bl)

#define LIS_QUAD_FMA_SSE2_ALIGN(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				sh  = _mm_set_pd(SPLITTER,SPLITTER); \
				ch = _mm_set_pd((b_hi),(c_hi)); \
				bh  = _mm_unpackhi_pd(ch,ch); \
				cl = _mm_set_pd((c_lo),(b_lo)); \
				bh  = _mm_mul_sd(bh,ch); \
				sh  = _mm_mul_pd(sh,ch); \
				th  = _mm_sub_pd(sh,ch); \
				sh  = _mm_sub_pd(sh,th); \
				th  = _mm_mul_pd(ch,cl); \
				ch = _mm_sub_pd(ch,sh); \
				eh  = _mm_unpackhi_pd(sh,ch); \
				wh  = _mm_unpacklo_pd(sh,ch); \
				ch = _mm_unpackhi_pd(ch,wh); \
				wh  = _mm_mul_pd(wh,eh); \
				sh  = _mm_mul_pd(sh,ch); \
				ch = _mm_unpackhi_pd(sh,sh); \
				cl = _mm_unpackhi_pd(wh,wh); \
				eh  = _mm_unpackhi_pd(th,th); \
				wh  = _mm_sub_sd(wh,bh); \
				wh  = _mm_add_sd(wh,ch); \
				wh  = _mm_add_sd(wh,sh); \
				wh  = _mm_add_sd(wh,cl); \
				wh  = _mm_add_sd(wh,eh); \
				wh  = _mm_add_sd(wh,th); \
				sh  = _mm_add_sd(wh,bh); \
				th  = _mm_sub_sd(sh,bh); \
				wh  = _mm_sub_sd(wh,th); \
				bl  = _mm_load_pd(&(a_hi)); \
				sh  = _mm_unpacklo_pd(sh,wh); \
				t0 = _mm_add_pd(bl,sh); \
				eh = _mm_sub_pd(t0,bl); \
				th = _mm_sub_pd(t0,eh); \
				sh = _mm_sub_pd(sh,eh); \
				bl = _mm_sub_pd(bl,th); \
				bl = _mm_add_pd(bl,sh); \
				eh = _mm_unpackhi_pd(bl,bl); \
				sh = _mm_unpackhi_pd(t0,t0); \
				th = t0; \
				bl = _mm_add_sd(bl,sh); \
				th = _mm_add_sd(th,bl); \
				t0 = _mm_sub_sd(th,t0); \
				bl = _mm_sub_sd(bl,t0); \
				bl = _mm_add_sd(bl,eh); \
				sh = _mm_add_sd(th,bl); \
				_mm_store_sd(&(d_hi),sh); \
				sh = _mm_sub_sd(sh,th); \
				bl = _mm_sub_sd(bl,sh); \
				_mm_store_sd(&(d_lo),bl)

#define LIS_QUAD_FMAD_SSE2(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c) \
				sh  = _mm_set_pd(SPLITTER,SPLITTER); \
				ch = _mm_set_pd((b_hi),(c)); \
				bh  = _mm_unpackhi_pd(ch,ch); \
				cl = _mm_load_sd(&(b_lo)); \
				bh  = _mm_mul_sd(bh,ch); \
				cl = _mm_mul_sd(cl,ch); \
				sh  = _mm_mul_pd(sh,ch); \
				th  = _mm_sub_pd(sh,ch); \
				sh  = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,sh); \
				eh  = _mm_unpackhi_pd(sh,ch); \
				wh  = _mm_unpacklo_pd(sh,ch); \
				ch = _mm_unpackhi_pd(ch,wh);  \
				wh  = _mm_mul_pd(wh,eh); \
				sh  = _mm_mul_pd(sh,ch); \
				ch = _mm_unpackhi_pd(sh,sh); \
				th  = _mm_unpackhi_pd(wh,wh); \
				wh  = _mm_sub_sd(wh,bh); \
				wh  = _mm_add_sd(wh,ch); \
				wh  = _mm_add_sd(wh,sh); \
				wh  = _mm_add_sd(wh,th); \
				wh  = _mm_add_sd(wh,cl); \
				sh  = _mm_add_sd(wh,bh); \
				th  = _mm_sub_sd(sh,bh); \
				wh  = _mm_sub_sd(wh,th); \
				sh  = _mm_unpacklo_pd(sh,wh); \
				bl = _mm_set_pd((a_lo),(a_hi)); \
				t0 = _mm_add_pd(bl,sh); \
				eh = _mm_sub_pd(t0,bl); \
				th = _mm_sub_pd(t0,eh); \
				sh = _mm_sub_pd(sh,eh); \
				bl = _mm_sub_pd(bl,th); \
				bl = _mm_add_pd(bl,sh); \
				eh = _mm_unpackhi_pd(bl,bl); \
				sh = _mm_unpackhi_pd(t0,t0); \
				th = t0; \
				bl = _mm_add_sd(bl,sh); \
				th = _mm_add_sd(th,bl); \
				t0 = _mm_sub_sd(th,t0); \
				bl = _mm_sub_sd(bl,t0); \
				bl = _mm_add_sd(bl,eh); \
				sh = _mm_add_sd(th,bl); \
				_mm_store_sd(&(d_hi),sh); \
				sh = _mm_sub_sd(sh,th); \
				bl = _mm_sub_sd(bl,sh); \
				_mm_store_sd(&(d_lo),bl)

#define LIS_QUAD_FMAD_SSE2_ALIGN(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c) \
				sh  = _mm_set_pd(SPLITTER,SPLITTER); \
				ch = _mm_set_pd((b_hi),(c)); \
				bh  = _mm_unpackhi_pd(ch,ch); \
				cl = _mm_load_sd(&(b_lo)); \
				bh  = _mm_mul_sd(bh,ch); \
				cl = _mm_mul_sd(cl,ch); \
				sh  = _mm_mul_pd(sh,ch); \
				th  = _mm_sub_pd(sh,ch); \
				sh  = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,sh); \
				eh  = _mm_unpackhi_pd(sh,ch); \
				wh  = _mm_unpacklo_pd(sh,ch); \
				ch = _mm_unpackhi_pd(ch,wh);  \
				wh  = _mm_mul_pd(wh,eh); \
				sh  = _mm_mul_pd(sh,ch); \
				ch = _mm_unpackhi_pd(sh,sh); \
				th  = _mm_unpackhi_pd(wh,wh); \
				wh  = _mm_sub_sd(wh,bh); \
				wh  = _mm_add_sd(wh,ch); \
				wh  = _mm_add_sd(wh,sh); \
				wh  = _mm_add_sd(wh,th); \
				wh  = _mm_add_sd(wh,cl); \
				sh  = _mm_add_sd(wh,bh); \
				th  = _mm_sub_sd(sh,bh); \
				wh  = _mm_sub_sd(wh,th); \
				sh  = _mm_unpacklo_pd(sh,wh); \
				bl  = _mm_load_pd(&(a_hi)); \
				t0 = _mm_add_pd(bl,sh); \
				eh = _mm_sub_pd(t0,bl); \
				th = _mm_sub_pd(t0,eh); \
				sh = _mm_sub_pd(sh,eh); \
				bl = _mm_sub_pd(bl,th); \
				bl = _mm_add_pd(bl,sh); \
				eh = _mm_unpackhi_pd(bl,bl); \
				sh = _mm_unpackhi_pd(t0,t0); \
				th = t0; \
				bl = _mm_add_sd(bl,sh); \
				th = _mm_add_sd(th,bl); \
				t0 = _mm_sub_sd(th,t0); \
				bl = _mm_sub_sd(bl,t0); \
				bl = _mm_add_sd(bl,eh); \
				sh = _mm_add_sd(th,bl); \
				_mm_store_sd(&(d_hi),sh); \
				sh = _mm_sub_sd(sh,th); \
				bl = _mm_sub_sd(bl,sh); \
				_mm_store_sd(&(d_lo),bl)

#define LIS_QUAD_FMA2_SSE2(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				t0  = _mm_set_pd(SPLITTER,SPLITTER); \
				bh = _mm_load_pd((b_hi)); \
				ch = _mm_loadu_pd(&(c_hi)); \
				bl = _mm_load_pd((b_lo)); \
				cl = _mm_loadu_pd(&(c_lo)); \
				p1  = _mm_mul_pd(bh,ch); \
				sh  = _mm_mul_pd(t0,bh); \
				sl  = _mm_mul_pd(t0,ch); \
				th  = _mm_sub_pd(sh,bh); \
				tl  = _mm_sub_pd(sl,ch); \
				sh  = _mm_sub_pd(sh,th); \
				sl  = _mm_sub_pd(sl,tl); \
				t1  = _mm_mul_pd(bh,cl); \
				wh  = _mm_sub_pd(bh,sh); \
				t2  = _mm_mul_pd(ch,bl); \
				wl  = _mm_sub_pd(ch,sl); \
				t0  = _mm_mul_pd(wh,wl); \
				p2  = _mm_mul_pd(sh,sl); \
				sh  = _mm_mul_pd(sh,wl); \
				sl  = _mm_mul_pd(sl,wh); \
				p2  = _mm_sub_pd(p2,p1); \
				p2  = _mm_add_pd(p2,sh); \
				p2  = _mm_add_pd(p2,sl); \
				p2  = _mm_add_pd(p2,t0); \
				p2  = _mm_add_pd(p2,t1); \
				p2  = _mm_add_pd(p2,t2); \
				ch  = _mm_add_pd(p1,p2); \
				t1  = _mm_sub_pd(ch,p1); \
				p2  = _mm_sub_pd(p2,t1); \
				bh = _mm_loadu_pd(&(a_hi)); \
				bl = _mm_loadu_pd(&(a_lo)); \
				sh = _mm_add_pd(bh,ch); \
				th = _mm_sub_pd(sh,bh); \
				t0 = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,th); \
				bh = _mm_sub_pd(bh,t0); \
				bh = _mm_add_pd(bh,ch); \
				sl = _mm_add_pd(bl,p2); \
				th = _mm_sub_pd(sl,bl); \
				t0 = _mm_sub_pd(sl,th); \
				p2 = _mm_sub_pd(p2,th); \
				bl = _mm_sub_pd(bl,t0); \
				bl = _mm_add_pd(bl,p2); \
				bh = _mm_add_pd(bh,sl); \
				th = sh; \
				th = _mm_add_pd(th,bh); \
				sh = _mm_sub_pd(th,sh); \
				bh = _mm_sub_pd(bh,sh); \
				bh = _mm_add_pd(bh,bl); \
				sh = _mm_add_pd(th,bh); \
				_mm_storeu_pd(&(d_hi),sh); \
				sh = _mm_sub_pd(sh,th); \
				bh = _mm_sub_pd(bh,sh); \
				_mm_storeu_pd(&(d_lo),bh)


#define LIS_QUAD_FMA2_SSE2_ALIGN(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				t0  = _mm_set_pd(SPLITTER,SPLITTER); \
				bh = _mm_loadu_pd(&(b_hi)); \
				ch = _mm_loadu_pd(&(c_hi)); \
				bl = _mm_loadu_pd(&(b_lo)); \
				cl = _mm_loadu_pd(&(c_lo)); \
				p1  = _mm_mul_pd(bh,ch); \
				sh  = _mm_mul_pd(t0,bh); \
				sl  = _mm_mul_pd(t0,ch); \
				th  = _mm_sub_pd(sh,bh); \
				tl  = _mm_sub_pd(sl,ch); \
				sh  = _mm_sub_pd(sh,th); \
				sl  = _mm_sub_pd(sl,tl); \
				t1  = _mm_mul_pd(bh,cl); \
				wh  = _mm_sub_pd(bh,sh); \
				t2  = _mm_mul_pd(ch,bl); \
				wl  = _mm_sub_pd(ch,sl); \
				t0  = _mm_mul_pd(wh,wl); \
				p2  = _mm_mul_pd(sh,sl); \
				sh  = _mm_mul_pd(sh,wl); \
				sl  = _mm_mul_pd(sl,wh); \
				p2  = _mm_sub_pd(p2,p1); \
				p2  = _mm_add_pd(p2,sh); \
				p2  = _mm_add_pd(p2,sl); \
				p2  = _mm_add_pd(p2,t0); \
				p2  = _mm_add_pd(p2,t1); \
				p2  = _mm_add_pd(p2,t2); \
				ch  = _mm_add_pd(p1,p2); \
				t1  = _mm_sub_pd(ch,p1); \
				p2  = _mm_sub_pd(p2,t1); \
				bh = _mm_load_pd(&(a_hi)); \
				bl = _mm_load_pd(&(a_lo)); \
				sh = _mm_add_pd(bh,ch); \
				th = _mm_sub_pd(sh,bh); \
				t0 = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,th); \
				bh = _mm_sub_pd(bh,t0); \
				bh = _mm_add_pd(bh,ch); \
				sl = _mm_add_pd(bl,p2); \
				th = _mm_sub_pd(sl,bl); \
				t0 = _mm_sub_pd(sl,th); \
				p2 = _mm_sub_pd(p2,th); \
				bl = _mm_sub_pd(bl,t0); \
				bl = _mm_add_pd(bl,p2); \
				bh = _mm_add_pd(bh,sl); \
				th = sh; \
				th = _mm_add_pd(th,bh); \
				sh = _mm_sub_pd(th,sh); \
				bh = _mm_sub_pd(bh,sh); \
				bh = _mm_add_pd(bh,bl); \
				sh = _mm_add_pd(th,bh); \
				_mm_store_pd(&(d_hi),sh); \
				sh = _mm_sub_pd(sh,th); \
				bh = _mm_sub_pd(bh,sh); \
				_mm_store_pd(&(d_lo),bh)

#define LIS_QUAD_FMAD2_SSE2_ALIGN2(a1_hi,a1_lo,a2_hi,a2_lo,b_hi,b_lo,c) \
				t0  = _mm_set_pd(SPLITTER,SPLITTER); \
				bh  = _mm_load_pd(&(b_hi)); \
				bl  = _mm_load_pd(&(b_lo)); \
				ch  = _mm_loadu_pd(&(c)); \
				p1  = _mm_mul_pd(bh,ch); \
				bl  = _mm_mul_pd(bl,ch); \
				sh  = _mm_mul_pd(t0,bh); \
				th  = _mm_sub_pd(sh,bh); \
				sh  = _mm_sub_pd(sh,th); \
				bh  = _mm_sub_pd(bh,sh); \
				sl  = _mm_mul_pd(t0,ch); \
				tl  = _mm_sub_pd(sl,ch); \
				sl  = _mm_sub_pd(sl,tl); \
				ch  = _mm_sub_pd(ch,sl); \
				t2  = _mm_mul_pd(bh,ch); \
				p2  = _mm_mul_pd(sh,sl); \
				t0  = _mm_mul_pd(sh,ch); \
				t1  = _mm_mul_pd(sl,bh); \
				p2  = _mm_sub_pd(p2,p1); \
				p2  = _mm_add_pd(p2,t0); \
				p2  = _mm_add_pd(p2,t1); \
				p2  = _mm_add_pd(p2,t2); \
				p2  = _mm_add_pd(p2,bl); \
				ch  = _mm_add_pd(p1,p2); \
				t1  = _mm_sub_pd(ch,p1); \
				p2  = _mm_sub_pd(p2,t1); \
				bh  = _mm_set_pd((a2_hi),(a1_hi)); \
				bl  = _mm_set_pd((a2_lo),(a1_lo)); \
				sh = _mm_add_pd(bh,ch); \
				th = _mm_sub_pd(sh,bh); \
				t0 = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,th); \
				bh = _mm_sub_pd(bh,t0); \
				bh = _mm_add_pd(bh,ch); \
				sl = _mm_add_pd(bl,p2); \
				th = _mm_sub_pd(sl,bl); \
				t0 = _mm_sub_pd(sl,th); \
				p2 = _mm_sub_pd(p2,th); \
				bl = _mm_sub_pd(bl,t0); \
				bl = _mm_add_pd(bl,p2); \
				bh = _mm_add_pd(bh,sl); \
				th = sh; \
				th = _mm_add_pd(th,bh); \
				sh = _mm_sub_pd(th,sh); \
				bh = _mm_sub_pd(bh,sh); \
				bh = _mm_add_pd(bh,bl); \
				sh = _mm_add_pd(th,bh); \
				_mm_storel_pd(&(a1_hi),sh); \
				_mm_storeh_pd(&(a2_hi),sh); \
				sh = _mm_sub_pd(sh,th); \
				bh = _mm_sub_pd(bh,sh); \
				_mm_storel_pd(&(a1_lo),bh); \
				_mm_storeh_pd(&(a2_lo),bh);

#define LIS_QUAD_FMAD2_SSE2_ALIGN(d_hi,d_lo,a_hi,a_lo,b1_hi,b1_lo,b2_hi,b2_lo,c) \
				t0  = _mm_set_pd(SPLITTER,SPLITTER); \
				bh  = _mm_set_pd((b2_hi),(b1_hi)); \
				bl  = _mm_set_pd((b2_lo),(b1_lo)); \
				ch  = _mm_loadu_pd(&(c)); \
				p1  = _mm_mul_pd(bh,ch); \
				bl  = _mm_mul_pd(bl,ch); \
				sh  = _mm_mul_pd(t0,bh); \
				th  = _mm_sub_pd(sh,bh); \
				sh  = _mm_sub_pd(sh,th); \
				bh  = _mm_sub_pd(bh,sh); \
				sl  = _mm_mul_pd(t0,ch); \
				tl  = _mm_sub_pd(sl,ch); \
				sl  = _mm_sub_pd(sl,tl); \
				ch  = _mm_sub_pd(ch,sl); \
				t2  = _mm_mul_pd(bh,ch); \
				p2  = _mm_mul_pd(sh,sl); \
				t0  = _mm_mul_pd(sh,ch); \
				t1  = _mm_mul_pd(sl,bh); \
				p2  = _mm_sub_pd(p2,p1); \
				p2  = _mm_add_pd(p2,t0); \
				p2  = _mm_add_pd(p2,t1); \
				p2  = _mm_add_pd(p2,t2); \
				p2  = _mm_add_pd(p2,bl); \
				ch  = _mm_add_pd(p1,p2); \
				t1  = _mm_sub_pd(ch,p1); \
				p2  = _mm_sub_pd(p2,t1); \
				bh = _mm_load_pd(&(a_hi)); \
				bl = _mm_load_pd(&(a_lo)); \
				sh = _mm_add_pd(bh,ch); \
				th = _mm_sub_pd(sh,bh); \
				t0 = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,th); \
				bh = _mm_sub_pd(bh,t0); \
				bh = _mm_add_pd(bh,ch); \
				sl = _mm_add_pd(bl,p2); \
				th = _mm_sub_pd(sl,bl); \
				t0 = _mm_sub_pd(sl,th); \
				p2 = _mm_sub_pd(p2,th); \
				bl = _mm_sub_pd(bl,t0); \
				bl = _mm_add_pd(bl,p2); \
				bh = _mm_add_pd(bh,sl); \
				th = sh; \
				th = _mm_add_pd(th,bh); \
				sh = _mm_sub_pd(th,sh); \
				bh = _mm_sub_pd(bh,sh); \
				bh = _mm_add_pd(bh,bl); \
				sh = _mm_add_pd(th,bh); \
				_mm_store_pd(&(d_hi),sh); \
				sh = _mm_sub_pd(sh,th); \
				bh = _mm_sub_pd(bh,sh); \
				_mm_store_pd(&(d_lo),bh)


/* d = a + b x b */
#define LIS_QUAD_FSA(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo) \
				p1  =  (b_hi) * (b_hi); \
				th  = SPLITTER * (b_hi); \
				bhi = th - (th-(b_hi)); \
				blo = (b_hi) - bhi; \
				p2  = ((bhi*bhi-p1)+2.0*bhi*blo)+blo*blo; \
				p2 += (2.0*(b_hi) * (b_lo)); \
				p2 += ((b_lo) * (b_lo)); \
				bhi = p1 + p2; \
				p2  = p2  - (bhi - p1); \
				sh  = (a_hi) + bhi; \
				th  = sh - (a_hi); \
				eh  = ((a_hi) - (sh - th)) + (bhi - th); \
				sl  = (a_lo) + p2; \
				tl  = sl - (a_lo); \
				el  = ((a_lo) - (sl - tl)) + (p2 - tl); \
				eh += sl; \
				th  = sh + eh; \
				eh  = eh - (th - sh); \
				eh += el; \
				(d_hi) = th + eh; \
				(d_lo) = eh  - ((d_hi) - th)

#define LIS_QUAD_FSA_SSE2_ALIGN(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo) \
				sh  = _mm_set_sd(SPLITTER); \
				bh  = _mm_load_sd(&(b_hi)); \
				bl  = _mm_load_sd(&(b_lo)); \
				p1  = _mm_mul_sd(bh,bh); \
				sh  = _mm_mul_sd(sh,bh); \
				wh  = _mm_sub_sd(sh,bh); \
				sh  = _mm_sub_sd(sh,wh); \
				cl  = _mm_sub_sd(bh,sh); \
				wh  = _mm_mul_sd(sh,sh); \
				sh  = _mm_add_sd(sh,sh); \
				sh  = _mm_mul_sd(sh,cl); \
				cl  = _mm_mul_sd(cl,cl); \
				wh  = _mm_sub_sd(wh,p1); \
				wh  = _mm_add_sd(wh,sh); \
				wh  = _mm_add_sd(wh,cl); \
				bh  = _mm_add_sd(bh,bh); \
				bh  = _mm_mul_sd(bh,bl); \
				bl  = _mm_mul_sd(bl,bl); \
				wh  = _mm_add_sd(wh,bh); \
				wh  = _mm_add_sd(wh,bl); \
				sh  = _mm_add_sd(p1,wh); \
				cl  = _mm_sub_sd(sh,p1); \
				wh  = _mm_sub_sd(wh,cl); \
				sh  = _mm_unpacklo_pd(sh,wh); \
				bl  = _mm_load_pd(&(a_hi)); \
				t0 = _mm_add_pd(bl,sh); \
				eh = _mm_sub_pd(t0,bl); \
				th = _mm_sub_pd(t0,eh); \
				sh = _mm_sub_pd(sh,eh); \
				bl = _mm_sub_pd(bl,th); \
				bl = _mm_add_pd(bl,sh); \
				eh = _mm_unpackhi_pd(bl,bl); \
				sh = _mm_unpackhi_pd(t0,t0); \
				th = t0; \
				bl = _mm_add_sd(bl,sh); \
				th = _mm_add_sd(th,bl); \
				t0 = _mm_sub_sd(th,t0); \
				bl = _mm_sub_sd(bl,t0); \
				bl = _mm_add_sd(bl,eh); \
				sh = _mm_add_sd(th,bl); \
				_mm_store_sd(&(d_hi),sh); \
				sh = _mm_sub_sd(sh,th); \
				bl = _mm_sub_sd(bl,sh); \
				_mm_store_sd(&(d_lo),bl)

#define LIS_QUAD_FSA2_SSE2_ALIGN(d_hi,d_lo,a_hi,a_lo,b_hi,b_lo) \
				ch  = _mm_set_pd(SPLITTER,SPLITTER); \
				bh = _mm_loadu_pd(&(b_hi)); \
				bl = _mm_loadu_pd(&(b_lo)); \
				p1  = _mm_mul_pd(bh,bh); \
				ch  = _mm_mul_pd(ch,bh); \
				p2  = _mm_sub_pd(ch,bh); \
				ch  = _mm_sub_pd(ch,p2); \
				cl  = _mm_sub_pd(bh,ch); \
				p2  = _mm_mul_pd(ch,ch); \
				ch  = _mm_add_pd(ch,ch); \
				ch  = _mm_mul_pd(ch,cl); \
				cl  = _mm_mul_pd(cl,cl); \
				p2  = _mm_sub_pd(p2,p1); \
				p2  = _mm_add_pd(p2,ch); \
				p2  = _mm_add_pd(p2,cl); \
				bh  = _mm_add_pd(bh,bh); \
				bh  = _mm_mul_pd(bh,bl); \
				bl  = _mm_mul_pd(bl,bl); \
				p2  = _mm_add_pd(p2,bh); \
				p2  = _mm_add_pd(p2,bl); \
				ch  = _mm_add_pd(p1,p2); \
				cl  = _mm_sub_pd(ch,p1); \
				p2  = _mm_sub_pd(p2,cl); \
				bh = _mm_load_pd(&(a_hi)); \
				bl = _mm_load_pd(&(a_lo)); \
				sh = _mm_add_pd(bh,ch); \
				th = _mm_sub_pd(sh,bh); \
				t0 = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,th); \
				bh = _mm_sub_pd(bh,t0); \
				bh = _mm_add_pd(bh,ch); \
				sl = _mm_add_pd(bl,p2); \
				th = _mm_sub_pd(sl,bl); \
				t0 = _mm_sub_pd(sl,th); \
				p2 = _mm_sub_pd(p2,th); \
				bl = _mm_sub_pd(bl,t0); \
				bl = _mm_add_pd(bl,p2); \
				bh = _mm_add_pd(bh,sl); \
				th = sh; \
				th = _mm_add_pd(th,bh); \
				sh = _mm_sub_pd(th,sh); \
				bh = _mm_sub_pd(bh,sh); \
				bh = _mm_add_pd(bh,bl); \
				sh = _mm_add_pd(th,bh); \
				_mm_store_pd(&(d_hi),sh); \
				sh = _mm_sub_pd(sh,th); \
				bh = _mm_sub_pd(bh,sh); \
				_mm_store_pd(&(d_lo),bh)

/* a = b + c */
#define LIS_QUAD_ADD(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				sh =  (b_hi) + (c_hi); \
				th =  sh - (b_hi); \
				eh =  ((b_hi) - (sh - th)) + ((c_hi) - th); \
				sl =  (b_lo) + (c_lo); \
				tl =  sl - (b_lo); \
				el =  ((b_lo) - (sl - tl)) + ((c_lo) - tl); \
				eh += sl; \
				th = sh + eh; \
				eh = eh - (th - sh); \
				eh += el; \
				(a_hi) = th + eh; \
				(a_lo) = eh  - ((a_hi) - th)

#define LIS_QUAD_ADD_SSE2(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				sh = _mm_set_pd((b_lo),(b_hi)); \
				bl = _mm_set_pd((c_lo),(c_hi)); \
				t0 = _mm_add_pd(bl,sh); \
				eh = _mm_sub_pd(t0,bl); \
				th = _mm_sub_pd(t0,eh); \
				sh = _mm_sub_pd(sh,eh); \
				bl = _mm_sub_pd(bl,th); \
				bl = _mm_add_pd(bl,sh); \
				eh = _mm_unpackhi_pd(bl,bl); \
				sh = _mm_unpackhi_pd(t0,t0); \
				th = t0; \
				bl = _mm_add_sd(bl,sh); \
				th = _mm_add_sd(th,bl); \
				t0 = _mm_sub_sd(th,t0); \
				bl = _mm_sub_sd(bl,t0); \
				bl = _mm_add_sd(bl,eh); \
				sh = _mm_add_sd(th,bl); \
				_mm_store_sd(&(a_hi),sh); \
				sh = _mm_sub_sd(sh,th); \
				bl = _mm_sub_sd(bl,sh); \
				_mm_store_sd(&(a_lo),bl)

#define LIS_QUAD_ADD_SSE2_ALIGN(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				bl  = _mm_load_pd((b_hi)); \
				sh  = _mm_load_pd((c_hi)); \
				t0 = _mm_add_pd(bl,sh); \
				eh = _mm_sub_pd(t0,bl); \
				th = _mm_sub_pd(t0,eh); \
				sh = _mm_sub_pd(sh,eh); \
				bl = _mm_sub_pd(bl,th); \
				bl = _mm_add_pd(bl,sh); \
				eh = _mm_unpackhi_pd(bl,bl); \
				sh = _mm_unpackhi_pd(t0,t0); \
				th = t0; \
				bl = _mm_add_sd(bl,sh); \
				th = _mm_add_sd(th,bl); \
				t0 = _mm_sub_sd(th,t0); \
				bl = _mm_sub_sd(bl,t0); \
				bl = _mm_add_sd(bl,eh); \
				sh = _mm_add_sd(th,bl); \
				_mm_store_sd(&(a_hi),sh); \
				sh = _mm_sub_sd(sh,th); \
				bl = _mm_sub_sd(bl,sh); \
				_mm_store_sd(&(a_lo),bl)

/* a = b x c */
#define LIS_QUAD_MUL(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				p1  = (b_hi) * (c_hi); \
				t   = SPLITTER * (b_hi); \
				bhi = t - (t-(b_hi)); \
				blo = (b_hi) - bhi; \
				t   = SPLITTER * (c_hi); \
				chi = t - (t-(c_hi)); \
				clo = (c_hi) - chi; \
				p2  = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo; \
				p2 += ((b_hi) * (c_lo)); \
				p2 += ((b_lo) * (c_hi)); \
				(a_hi) = p1 + p2; \
				(a_lo) = p2  - ((a_hi) - p1)

#define LIS_QUAD_MUL_SSE2(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				sh  = _mm_set_pd(SPLITTER,SPLITTER); \
				ch = _mm_set_pd((c_hi),(b_hi)); \
				bh  = _mm_unpackhi_pd(ch,ch); \
				tl = _mm_set_pd((b_lo),(c_lo)); \
				bh  = _mm_mul_sd(bh,ch); \
				sh  = _mm_mul_pd(sh,ch); \
				th  = _mm_sub_pd(sh,ch); \
				sh  = _mm_sub_pd(sh,th); \
				th  = _mm_mul_pd(ch,tl); \
				ch = _mm_sub_pd(ch,sh); \
				eh  = _mm_unpackhi_pd(sh,ch); \
				wh  = _mm_unpacklo_pd(sh,ch); \
				ch = _mm_unpackhi_pd(ch,wh);  \
				wh  = _mm_mul_pd(wh,eh); \
				sh  = _mm_mul_pd(sh,ch); \
				ch = _mm_unpackhi_pd(sh,sh); \
				tl = _mm_unpackhi_pd(wh,wh); \
				eh  = _mm_unpackhi_pd(th,th); \
				wh  = _mm_sub_sd(wh,bh); \
				wh  = _mm_add_sd(wh,ch); \
				wh  = _mm_add_sd(wh,sh); \
				wh  = _mm_add_sd(wh,tl); \
				wh  = _mm_add_sd(wh,eh); \
				wh  = _mm_add_sd(wh,th); \
				sh  = _mm_add_sd(wh,bh); \
				_mm_store_sd(&(a_hi),sh); \
				sh  = _mm_sub_sd(sh,bh); \
				wh  = _mm_sub_sd(wh,sh); \
				_mm_store_sd(&(a_lo),wh)

#define LIS_QUAD_MUL2_SSE2_ALIGN(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				t0  = _mm_set_pd(SPLITTER,SPLITTER); \
				bh = _mm_load_pd(&(c_hi)); \
				ch = _mm_loadu_pd(&(b_hi)); \
				bl = _mm_load_pd(&(c_hi)); \
				cl = _mm_loadu_pd(&(b_hi)); \
				p1  = _mm_mul_pd(bh,ch); \
				sh  = _mm_mul_pd(t0,bh); \
				sl  = _mm_mul_pd(t0,ch); \
				th  = _mm_sub_pd(sh,bh); \
				tl  = _mm_sub_pd(sl,ch); \
				sh  = _mm_sub_pd(sh,th); \
				sl  = _mm_sub_pd(sl,tl); \
				t1  = _mm_mul_pd(bh,cl); \
				wh  = _mm_sub_pd(bh,sh); \
				t2  = _mm_mul_pd(ch,bl); \
				wl  = _mm_sub_pd(ch,sl); \
				t0  = _mm_mul_pd(wh,wl); \
				p2  = _mm_mul_pd(sh,sl); \
				sh  = _mm_mul_pd(sh,wl); \
				sl  = _mm_mul_pd(sl,wh); \
				p2  = _mm_sub_pd(p2,p1); \
				p2  = _mm_add_pd(p2,sh); \
				p2  = _mm_add_pd(p2,sl); \
				p2  = _mm_add_pd(p2,t0); \
				p2  = _mm_add_pd(p2,t1); \
				p2  = _mm_add_pd(p2,t2); \
				ch  = _mm_add_pd(p1,p2); \
				_mm_storeu_pd(&(a_hi),ch); \
				ch  = _mm_sub_pd(ch,p1); \
				p2  = _mm_sub_pd(p2,ch); \
				_mm_storeu_pd(&(a_lo),p2)

/* a = b x c */
#define LIS_QUAD_MULD(a_hi,a_lo,b_hi,b_lo,c) \
				p1  = (b_hi) * (c); \
				t   = SPLITTER * (b_hi); \
				bhi = t - (t-(b_hi)); \
				blo = (b_hi) - bhi; \
				t   = SPLITTER * (c); \
				chi = t - (t-(c)); \
				clo = (c) - chi; \
				p2  = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo; \
				p2 += ((b_lo) * (c)); \
				(a_hi) = p1 + p2; \
				(a_lo) = p2  - ((a_hi) - p1)

#define LIS_QUAD_MULD_SSE2(a_hi,a_lo,b_hi,b_lo,c) \
				sh  = _mm_set_pd(SPLITTER,SPLITTER); \
				ch = _mm_set_pd((b_hi),(c)); \
				bh  = _mm_unpackhi_pd(ch,ch); \
				sl = _mm_load_sd(&(b_lo)); \
				bh  = _mm_mul_sd(bh,ch); \
				sl = _mm_mul_sd(sl,ch); \
				sh  = _mm_mul_pd(sh,ch); \
				th  = _mm_sub_pd(sh,ch); \
				sh  = _mm_sub_pd(sh,th); \
				ch = _mm_sub_pd(ch,sh); \
				t1  = _mm_unpackhi_pd(sh,ch); \
				t0  = _mm_unpacklo_pd(sh,ch); \
				ch = _mm_unpackhi_pd(ch,t0);  \
				t0  = _mm_mul_pd(t0,t1); \
				sh  = _mm_mul_pd(sh,ch); \
				ch = _mm_unpackhi_pd(sh,sh); \
				th  = _mm_unpackhi_pd(t0,t0); \
				t0  = _mm_sub_sd(t0,bh); \
				t0  = _mm_add_sd(t0,ch); \
				t0  = _mm_add_sd(t0,sh); \
				t0  = _mm_add_sd(t0,th); \
				t0  = _mm_add_sd(t0,sl); \
				th  = _mm_add_sd(t0,bh); \
				_mm_store_sd(&(a_hi),th); \
				th  = _mm_sub_sd(th,bh); \
				t0  = _mm_sub_sd(t0,th); \
				_mm_store_sd(&(a_lo),t0)

#define LIS_QUAD_MULD2_SSE2(a_hi,a_lo,b_hi,b_lo,c) \
				t0  = _mm_set_pd(SPLITTER,SPLITTER); \
				bh  = _mm_loadu_pd(&(b_hi)); \
				bl  = _mm_loadu_pd(&(b_hi)); \
				ch  = _mm_load_pd(&(c)); \
				p1  = _mm_mul_pd(bh,ch); \
				bl  = _mm_mul_pd(bl,ch); \
				sh  = _mm_mul_pd(t0,bh); \
				th  = _mm_sub_pd(sh,bh); \
				sh  = _mm_sub_pd(sh,th); \
				bh  = _mm_sub_pd(bh,sh); \
				sl  = _mm_mul_pd(t0,ch); \
				tl  = _mm_sub_pd(sl,ch); \
				sl  = _mm_sub_pd(sl,tl); \
				ch  = _mm_sub_pd(ch,sl); \
				t2  = _mm_mul_pd(bh,ch); \
				p2  = _mm_mul_pd(sh,sl); \
				t0  = _mm_mul_pd(sh,ch); \
				t1  = _mm_mul_pd(sl,bh); \
				p2  = _mm_sub_pd(p2,p1); \
				p2  = _mm_add_pd(p2,t0); \
				p2  = _mm_add_pd(p2,t1); \
				p2  = _mm_add_pd(p2,t2); \
				p2  = _mm_add_pd(p2,bl); \
				t0  = _mm_add_pd(p1,p2); \
				_mm_storeu_pd(&(a_hi),t0); \
				t0  = _mm_sub_pd(t0,p1); \
				p2  = _mm_sub_pd(p2,t0); \
				_mm_storeu_pd(&(a_lo),p2)

/* a = b / c */
#define LIS_QUAD_DIV(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				q1  = (b_hi) / (c_hi); \
				p1 =  (c_hi) * q1; \
				t   = SPLITTER * (c_hi); \
				bhi = t - (t-(c_hi)); \
				blo = (c_hi) - bhi; \
				t   = SPLITTER * q1; \
				chi = t - (t-q1); \
				clo = q1 - chi; \
				p2 = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo; \
				p2 += ((c_lo) * q1); \
				r.hi = p1 + p2; \
				r.lo = p2  - (r.hi - p1); \
				s1 =  (b_hi) - r.hi; \
				t =  s1 - (b_hi); \
				s2 =  ((b_hi) - (s1 - t)) - (r.hi + t); \
				s2 -= r.lo; \
				s2 += (b_lo); \
				q2  = (s1+s2) / (c_hi); \
				(a_hi) = q1 + q2; \
				(a_lo) = q2  - ((a_hi) - q1)

#define LIS_QUAD_DIV_SSE2(a_hi,a_lo,b_hi,b_lo,c_hi,c_lo) \
				xs  = _mm_set_pd(SPLITTER,SPLITTER); \
				xb  = _mm_set_pd((b_lo),(b_hi)); \
				xc  = _mm_set_pd((c_lo),(c_hi)); \
				xq  = xb; \
				xw  = xc; \
				xq  = _mm_div_sd(xq,xc); \
				xw  = _mm_unpacklo_pd(xw,xq); \
				xc  = _mm_move_sd(xc,xq); \
				xq  = xw; \
				xs  = _mm_mul_pd(xs,xw); \
				xt  = _mm_sub_pd(xs,xw); \
				xs  = _mm_sub_pd(xs,xt); \
				xc  = _mm_mul_pd(xc,xw); \
				xw  = _mm_sub_pd(xw,xs); \
				xt  = xs; \
				xr  = xw; \
				xt  = _mm_unpackhi_pd(xt,xt); \
				xr  = _mm_unpackhi_pd(xr,xr); \
				xe  = xt; \
				xt  = _mm_mul_sd(xt,xw); \
				xe  = _mm_mul_sd(xe,xs); \
				xw  = _mm_mul_sd(xw,xr); \
				xs  = _mm_mul_sd(xs,xr); \
				xr  = xc; \
				xr  = _mm_unpackhi_pd(xr,xr); \
				xe  = _mm_sub_sd(xe,xc); \
				xe  = _mm_add_sd(xe,xs); \
				xe  = _mm_add_sd(xe,xt); \
				xe  = _mm_add_sd(xe,xw); \
				xe  = _mm_add_sd(xe,xr); \
				xr  = xc; \
				xr  = _mm_add_sd(xr,xe); \
				xw  = xr; \
				xw  = _mm_sub_sd(xw,xc); \
				xs  = xb; \
				xs  = _mm_sub_sd(xs,xr); \
				xe  = _mm_sub_sd(xe,xw); \
				xt  = xs; \
				xt  = _mm_sub_sd(xt,xb); \
				xr  = _mm_add_sd(xr,xt); \
				xw  = xs; \
				xw  = _mm_sub_sd(xw,xt); \
				xt  = xb; \
				xt  = _mm_unpackhi_pd(xt,xt); \
				xb  = _mm_sub_sd(xb,xw); \
				xb  = _mm_sub_sd(xb,xr); \
				xb  = _mm_sub_sd(xb,xe); \
				xb  = _mm_add_sd(xb,xt); \
				xt  = xq; \
				xt  = _mm_unpackhi_pd(xt,xt); \
				xe  = xt; \
				xs  = _mm_add_sd(xs,xb); \
				xs  = _mm_div_sd(xs,xq); \
				xt  = _mm_add_sd(xt,xs); \
				_mm_store_sd(&(a_hi),xt); \
				xt  = _mm_sub_sd(xt,xe); \
				xs  = _mm_sub_sd(xs,xt); \
				_mm_store_sd(&(a_lo),xs)


/* a = sqrt(b) */
#define LIS_QUAD_SQRT(a_hi,a_lo,b_hi,b_lo) \
				if( (b_hi)==0 ) \
				{ \
					(a_hi) = (a_lo) = 0.0; \
					return LIS_FAILS; \
				} \
				if( (b_hi)<0 ) \
				{ \
					printf("ERROR bh=%e\n",(b_hi)); \
					(a_hi) = (a_lo) = 0.0; \
					return LIS_FAILS; \
				} \
				p1  = 1.0 / sqrt((b_hi)); \
				p2 = (b_hi) * p1; \
				p1  = p1 * 0.5; \
				chi =  p2 * p2; \
				t   = SPLITTER * p2; \
				bhi  = t - (t-p2); \
				blo  = p2 - bhi; \
				clo = ((bhi*bhi-chi)+2.0*bhi*blo)+blo*blo; \
				sh  = (b_hi) - chi; \
				th  = sh - (b_hi); \
				eh  = ((b_hi) - (sh - th)) - (chi + th); \
				sl  = (b_lo) - clo; \
				tl  = sl - (b_lo); \
				el  = ((b_lo) - (sl - tl)) - (clo + tl); \
				eh += sl; \
				th  = sh + eh; \
				eh  = eh - (th - sh); \
				eh += el; \
				th  = th + eh; \
				p1  = p1 * th; \
				(a_hi) = p2 + p1; \
				(a_lo) = p1  - ((a_hi) - p2)

#define LIS_QUAD_SQRT_SSE2(a_hi,a_lo,b_hi,b_lo) \
				if( (b_hi)==0 ) \
				{ \
					(a_hi) = (a_lo) = 0.0; \
					return LIS_FAILS; \
				} \
				if( (b_hi)<0 ) \
				{ \
					printf("ERROR bh=%e\n",(b_hi)); \
					(a_hi) = (a_lo) = 0.0; \
					return LIS_FAILS; \
				} \
				wh  = _mm_set_sd(SPLITTER); \
				bh = _mm_load_pd(&(b_hi)); \
				bh = bh; \
				t0 = _mm_set_sd(1.0); \
				t1 = _mm_set_sd(0.5); \
				t2 = _mm_sqrt_pd(bh); \
				t0 = _mm_div_sd(t0,t2); \
				t2 = _mm_mul_sd(bh,t0); \
				t0 = _mm_mul_sd(t0,t1); \
				p1  = _mm_mul_sd(t2,t2); \
				wh  = _mm_mul_sd(wh,t2); \
				t1  = _mm_sub_sd(wh,t2); \
				wh  = _mm_sub_sd(wh,t1); \
				wl  = _mm_sub_sd(t2,wh); \
				t1  = _mm_mul_sd(wh,wh); \
				wh  = _mm_add_sd(wh,wh); \
				wh  = _mm_mul_sd(wh,wl); \
				wl  = _mm_mul_sd(wl,wl); \
				t1  = _mm_sub_sd(t1,p1); \
				t1  = _mm_add_sd(t1,wh); \
				t1  = _mm_add_sd(t1,wl); \
				p1 = _mm_unpacklo_pd(p1,t1); \
				sh = _mm_sub_pd(bh,p1); \
				eh = _mm_sub_pd(sh,bh); \
				th = _mm_sub_pd(sh,eh); \
				p1 = _mm_add_pd(p1,eh); \
				bh = _mm_sub_pd(bh,th); \
				bh = _mm_sub_pd(bh,p1); \
				eh = _mm_unpackhi_pd(bh,bh); \
				ch = _mm_unpackhi_pd(sh,sh); \
				th = sh; \
				bh = _mm_add_sd(bh,ch); \
				th = _mm_add_sd(th,bh); \
				sh = _mm_sub_sd(th,sh); \
				bh = _mm_sub_sd(bh,sh); \
				bh = _mm_add_sd(bh,eh); \
				th = _mm_add_sd(th,bh); \
				t0 = _mm_mul_sd(t0,th); \
				p1 = _mm_add_sd(t2,t0); \
				_mm_store_sd(&(a_hi),p1); \
				p1 = _mm_sub_sd(p1,t2); \
				t0 = _mm_sub_sd(t0,p1); \
				_mm_store_sd(&(a_lo),t0)

extern double *lis_quad_scalar_tmp;

#define LIS_QUAD_SCALAR_MALLOC(s,pos,num) \
				(s).hi = &lis_quad_scalar_tmp[2*(pos)]; \
				(s).lo = &lis_quad_scalar_tmp[2*(pos)+(num)]

#ifdef __cplusplus
extern "C"
{
#endif

extern void lisqp_minus(LIS_QDOUBLE *a);
extern void lisqp_zero(LIS_QDOUBLE *a);
extern void lisqp_one(LIS_QDOUBLE *a);
extern void lisqp_sqrt(LIS_QDOUBLE *a, const LIS_QDOUBLE *b);
extern void lisqp_npower(LIS_QDOUBLE *a, const LIS_QDOUBLE b, const int n);

extern void lisqp_add(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c);
extern void lisqp_add_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c);
extern void lisqp_add_d_d(LIS_QDOUBLE *a, const double b, const double c);

extern void lisqp_sub(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c);
extern void lisqp_sub_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c);

extern void lisqp_mul(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c);
extern void lisqp_mul_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c);
extern void lisqp_sqr(LIS_QDOUBLE *a, const LIS_QDOUBLE *b);

extern void lisqp_div(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c);
extern void lisqp_div_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c);
extern void lisqp_div_d_dd(LIS_QDOUBLE *a, const double b, const LIS_QDOUBLE *c);



extern void lisqp_minus_m(LIS_DOUBLE_DOUBLE *a);
extern void lisqp_zero_m(LIS_DOUBLE_DOUBLE *a, int n);
extern void lisqp_one_m(LIS_DOUBLE_DOUBLE *a);
extern int lisqp_malloc(LIS_DOUBLE_DOUBLE *a, int n);
extern int lisqp_free(LIS_DOUBLE_DOUBLE *a);


#ifdef __cplusplus
}
#endif

#endif
