/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#ifndef __LIS_SYSTEM_H__
#define __LIS_SYSTEM_H__

#include <stdarg.h>

#if defined(_DEBUG)
#define LIS_DEBUG_FUNC_IN	lis_debug_trace_func(1,__FUNC__)
#define LIS_DEBUG_FUNC_OUT	lis_debug_trace_func(0,__FUNC__)
#else
#define LIS_DEBUG_FUNC_IN
#define LIS_DEBUG_FUNC_OUT
#endif

#define LIS_SETERR(code,mess)			lis_error(__FILE__,__FUNC__,__LINE__,code,mess)
#define LIS_SETERR1(code,mess,a1)		lis_error(__FILE__,__FUNC__,__LINE__,code,mess,a1)
#define LIS_SETERR2(code,mess,a1,a2)	lis_error(__FILE__,__FUNC__,__LINE__,code,mess,a1,a2)
#define LIS_SETERR3(code,mess,a1,a2,a3)	lis_error(__FILE__,__FUNC__,__LINE__,code,mess,a1,a2,a3)
#define LIS_SETERR4(code,mess,a1,a2,a3,a4)	lis_error(__FILE__,__FUNC__,__LINE__,code,mess,a1,a2,a3,a4)
#define LIS_SETERR_MEM(size)			lis_error(__FILE__,__FUNC__,__LINE__,LIS_ERR_OUT_OF_MEMORY,"malloc size = %d\n",size)
#define LIS_SETERR_IMP					lis_error(__FILE__,__FUNC__,__LINE__,LIS_ERR_NOT_IMPLEMENTED,NULL)
#define LIS_SETERR_FIO					lis_error(__FILE__,__FUNC__,__LINE__,LIS_ERR_FILE_IO,"file i/o error\n")


typedef struct LIS_ARGS_STRUCT
{
  struct LIS_ARGS_STRUCT *next, *prev;
  char   *arg1;
  char   *arg2;
} *LIS_ARGS;

typedef struct LIS_HASH_STRUCT
{
	struct LIS_HASH_STRUCT *next;
	int		index;
	int		value;
} *LIS_HASH;

typedef struct LIS_HASH_STRUCT **LIS_HASHTABLE;



#ifdef __cplusplus
extern "C"
{
#endif
	extern LIS_ARGS cmd_args;
	extern LIS_SCALAR	*lis_vec_tmp;
	extern int lis_mpi_initialized;
#ifdef USE_MPI
	extern MPI_Op LIS_MPI_MSUM;
	extern MPI_Datatype LIS_MPI_MSCALAR;
#endif
	extern int lis_text2args(char *text, LIS_ARGS *args);
	extern int lis_arg2args(int argc, char *argv[], LIS_ARGS *args);
	extern int lis_args_free(LIS_ARGS args);
	extern int lis_debug_trace_func(int flag, char *func);
	extern void lis_debug_set_comm(LIS_Comm comm);
	extern int lis_printf(LIS_Comm comm, const char *mess, ...);
	extern int lis_error(const char *file, const char *func, const int line, const int code, const char *mess, ...);
	extern void *lis_malloc( size_t size);
	extern void *lis_calloc( size_t size);
	extern void *lis_realloc( void *p, size_t size );
	extern void lis_free(void *p);
	extern void lis_free2(int n, ...);
	extern int lis_is_malloc( void *p );
	extern void lis_sort_i(int is, int ie, int *i1);
	extern void lis_sort_ii(int is, int ie, int *i1, int *i2);
	extern void lis_sort_id(int is, int ie, int *i1, LIS_SCALAR *d1);
	extern void lis_sort_di(int is, int ie, LIS_SCALAR *d1, int *i1);
	extern void lis_sort_iid(int is, int ie, int *i1, int *i2, LIS_SCALAR *d1);
	extern void lis_sort_iiid(int is, int ie, int *i1, int *i2, int *i3, LIS_SCALAR *d1);
	extern void lis_sortr_ii(int is, int ie, int *i1, int *i2);
	extern void lis_sort_jds(int is, int ie, int maxnzr, int *i1, int *i2);
	extern int lis_bswap_int(int n, int *buf);
	extern int lis_bswap_double(int n, double *buf);	
	extern int lis_bswap_size_t(int n, size_t *buf);
	extern int lis_ranges_create(LIS_Comm comm, int *local_n, int *global_n, int **ranges, int *is, int *ie, int *nprocs, int *my_rank);
	extern int lis_hashtable_create(LIS_HASHTABLE *hashtable);
	extern int lis_hashtable_destroy(LIS_HASHTABLE hashtable);
	extern int lis_hashtable_clear(LIS_HASHTABLE hashtable);
	extern LIS_HASH lis_hashtable_search(LIS_HASHTABLE hashtable, int index);
	extern int lis_hashtable_set_value(LIS_HASHTABLE hashtable, int index, int value);
	extern int lis_hashtable_get_value(LIS_HASHTABLE hashtable, int index);

#ifdef __cplusplus
}
#endif

#endif
