/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#ifndef __LIS_VECOTR_H__
#define __LIS_VECTOR_H__

#define LIS_VECTOR_CHECK_ALL		0
#define LIS_VECTOR_CHECK_NULL		1
#define LIS_VECTOR_CHECK_ASSEMBLE	2



#ifdef __cplusplus
extern "C"
{
#endif

	extern int lis_vector_init(LIS_VECTOR *vec);
	extern int lis_vector_is_null(LIS_VECTOR v);
	extern int lis_vector_createex(int precision, LIS_Comm comm, LIS_VECTOR *vec);
	extern int lis_vector_reuse(LIS_VECTOR *vec);
	extern int lis_vector_duplicateex(int precision, void *A, LIS_VECTOR *vout);
	extern int lis_vector_unset(LIS_VECTOR vec);
	extern int lis_vector_set(LIS_VECTOR vec, LIS_SCALAR *value);
	extern int lis_vector_set_destroyflag(LIS_VECTOR v, int flag);
	extern int lis_vector_get_destroyflag(LIS_VECTOR v, int *flag);
	extern int lis_vector_check(LIS_VECTOR v, int level);
	/*******************/
	/* Operations      */
	/*******************/
	extern int lis_vector_axpyex(LIS_MSCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy);
	extern int lis_vector_xpayex(LIS_VECTOR vx, LIS_MSCALAR alpha, LIS_VECTOR vy);
	extern int lis_vector_axpyzex(LIS_MSCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy, LIS_VECTOR vz);
	extern int lis_vector_set_allex(LIS_MSCALAR alpha, LIS_VECTOR vx);
	extern int lis_vector_copyex(LIS_VECTOR vsrc, LIS_VECTOR vdst);
	extern int lis_vector_scaleex(LIS_MSCALAR alpha, LIS_VECTOR vx);
	extern int lis_vector_dotex(LIS_VECTOR vx, LIS_VECTOR vy, LIS_MSCALAR *val);
	extern int lis_vector_nrm2ex(LIS_VECTOR vx, LIS_MSCALAR *val);
	extern int lis_vector_sumex(LIS_VECTOR vx, LIS_MSCALAR *val);

	extern int lis_vector_axpyex_mmm(LIS_SCALAR_M alpha, LIS_VECTOR vx, LIS_VECTOR vy);
	extern int lis_vector_xpayex_mmm(LIS_VECTOR vx, LIS_SCALAR_M alpha, LIS_VECTOR vy);
	extern int lis_vector_axpyzex_mmmm(LIS_SCALAR_M alpha, LIS_VECTOR vx, LIS_VECTOR vy, LIS_VECTOR vz);
	extern int lis_vector_copyex_mm(LIS_VECTOR vx, LIS_VECTOR vy);
	extern int lis_vector_copyex_mn(LIS_VECTOR vx, LIS_VECTOR vy);
	extern int lis_vector_copyex_nm(LIS_VECTOR vx, LIS_VECTOR vy);
	extern int lis_vector_scaleex_mm(LIS_SCALAR_M alpha, LIS_VECTOR vx);
	extern int lis_vector_scaleex_nm(LIS_SCALAR alpha, LIS_VECTOR vx);
	extern int lis_vector_set_allex_nm(LIS_SCALAR alpha, LIS_VECTOR vx);
	extern int lis_vector_nrm2ex_mm(LIS_VECTOR vx, LIS_SCALAR_M *val);
	extern int lis_vector_dotex_mmm(LIS_VECTOR vx, LIS_VECTOR vy, LIS_SCALAR_M *val);
	extern int lis_vector_reciprocalex_m(LIS_VECTOR vx);

#ifdef __cplusplus
}
#endif

#endif
