MODULE hash_mod
#ifdef DEBUG
  INTEGER(kind=4) :: hash_conflict_count = 0 
#endif
    
  contains
    SUBROUTINE hash_zero(hash_size, hash_array, val, position, id)
    implicit none
    
    include 'precision.inc'
    INTEGER(kind=kint), intent(in)  :: hash_size, hash_array(:), val
    CHARACTER(len=*), intent(in) :: id
    INTEGER(kind=kint), intent(out) :: position

    INTEGER(kind=kint) :: i

    position = MOD(val, hash_size) + 1
    if(hash_array(position) /= 0) then
       position = MOD(val * MOD(val, 1001), hash_size) + 1
       if(hash_array(position) /= 0) then
#ifdef DEBUG          
          hash_conflict_count = hash_conflict_count + 1
#endif
          do i = position + 1, hash_size
             if(hash_array(i) == 0) exit
          end do
          if(i > hash_size)then
             do i = 1, position - 1
                if(hash_array(i) == 0) exit
             end do

#ifdef ALLOCATION_CHECK
             if(i >= position) then
                write(*,*) "enlarge ", id
                stop "error in hash_zero():"
             end if
#endif
          end if
          position = i
       end if
    end if
  END SUBROUTINE hash_zero

  SUBROUTINE hash(hash_size, hash_array, val, position, id)
    implicit none
    include 'precision.inc'
    
    INTEGER(kind=kint), intent(in)  :: hash_size, val, hash_array(:)
    CHARACTER(len=*), intent(in) :: id
    INTEGER(kind=kint), intent(out) :: position

    INTEGER(kind=4) :: i
    
    position = MOD(val, hash_size) + 1
    if(hash_array(position) /= 0 .and. hash_array(position) /= val) then
       position = MOD(val*MOD(val, 1001), hash_size) + 1

       if(hash_array(position) /= 0 .and. hash_array(position) /= val) then
#ifdef DEBUG          
          hash_conflict_count = hash_conflict_count + 1
#endif
          do i = position + 1, hash_size
             if(hash_array(i) == 0 .or. hash_array(i) == val) exit
          end do

          if(i > hash_size) then
             do i = 1, position - 1
                if(hash_array(i) == 0 .or. hash_array(i) == val) exit
             end do
#ifdef ALLOCATION_CHECK
             if(i >= position) then
                write(*,*) "enlarge ", id
                stop "error in hash_zero():"
             end if
#endif
          end if
          position = i
       end if
    end if
  END SUBROUTINE hash
  
END MODULE hash_mod
