/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_initialize_f
 * lis_finalize_f
 ************************************************/
#ifdef USE_FORTRAN

char	**f_argv_tmp;
int		f_argc_tmp;

#undef __FUNC__
#define __FUNC__ "lis_finitialize_f"
void lis_finitialize_f(int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	#ifdef USE_MPI
		MPI_Init(&f_argc_tmp, &f_argv_tmp);
	#endif

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_finalize_f"
void lis_finalize_f(int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_finalize();

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_wtime_f"
double lis_wtime_f(void)
{
	LIS_DEBUG_FUNC_IN;

	return lis_wtime();

	LIS_DEBUG_FUNC_OUT;
}

#undef __FUNC__
#define __FUNC__ "CHKERR_f"
void CHKERR_f(int *err)
{
	LIS_DEBUG_FUNC_IN;

	if(*err)
	{
		lis_finalize();
		exit(*err);
	}

	LIS_DEBUG_FUNC_OUT;
}


#undef __FUNC__
#define __FUNC__ "lis_set_argv_begin_f"
void lis_set_argv_begin_f(int *argc, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	f_argc_tmp = *argc+1;
	f_argv_tmp = (char **)lis_malloc(f_argc_tmp*sizeof(char *));
	if( f_argv_tmp==NULL )
	{
		LIS_SETERR_MEM(f_argc_tmp*sizeof(char *));
		*ierr = LIS_OUT_OF_MEMORY;
		return;
	}
	*ierr = LIS_SUCCESS;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_set_argv_f"
void lis_set_argv_f(int *no, char *argv, int *ierr, int len)
{
	int   i;
	char  *p;

	LIS_DEBUG_FUNC_IN;

	i = *no;
	f_argv_tmp[i] = (char *)lis_malloc((len+1)*sizeof(char));
	if( f_argv_tmp[i]==NULL )
	{
		LIS_SETERR_MEM((len+1)*sizeof(char));
		*ierr = LIS_OUT_OF_MEMORY;
		return;
	}
	memset(f_argv_tmp[i],0x20,(len+1)*sizeof(char));
	strncpy(f_argv_tmp[i],argv,len);
	p = &f_argv_tmp[i][len];
	if( len>0 )
	{
		while( *p==' ' ) p--;
		p++;
	}
	*p = '\0';
	*ierr = LIS_SUCCESS;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_set_argv_end_f"
void lis_set_argv_end_f(int *ierr)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	for(i=0;i<f_argc_tmp;i++)
	{
		lis_free(f_argv_tmp[i]);
	}
	lis_free(f_argv_tmp);
	f_argv_tmp = NULL;
	f_argc_tmp = 0;
	*ierr = LIS_SUCCESS;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_arg2args_f"
void lis_arg2args_f(int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_arg2args(f_argc_tmp,f_argv_tmp,&cmd_args);

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_input_f"
void lis_input_f(LIS_MATRIX_F *A, LIS_VECTOR_F *b, LIS_VECTOR_F *x, char *filename, int *ierr, int len)
{
	char		buf[1024];
	char		*p;

	LIS_DEBUG_FUNC_IN;

	memset(buf,0x20,sizeof(buf));
	strncpy(buf,filename,len);
	p = &buf[len];
	if( len>0 )
	{
		while( *p==' ' ) p--;
		p++;
	}
	*p = '\0';

	*ierr = lis_input((LIS_MATRIX)LIS_V2P(A),(LIS_VECTOR)LIS_V2P(b),(LIS_VECTOR)LIS_V2P(x),buf);

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_output_vector_f"
void lis_output_vector_f(LIS_VECTOR_F *v, int *format, char *filename, int *ierr, int len)
{
	char		buf[1024];
	char		*p;

	LIS_DEBUG_FUNC_IN;

	memset(buf,0x20,sizeof(buf));
	strncpy(buf,filename,len);
	p = &buf[len];
	if( len>0 )
	{
		while( *p==' ' ) p--;
		p++;
	}
	*p = '\0';

	*ierr = lis_output_vector((LIS_VECTOR)LIS_V2P(v),*format,buf);

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_output_residual_history_f"
void lis_output_residual_history_f(LIS_SOLVER_F *solver, char *filename, int *ierr, int len)
{
	char		buf[1024];
	char		*p;

	LIS_DEBUG_FUNC_IN;

	memset(buf,0x20,sizeof(buf));
	strncpy(buf,filename,len);
	p = &buf[len];
	if( len>0 )
	{
		while( *p==' ' ) p--;
		p++;
	}
	*p = '\0';

	*ierr = lis_output_residual_history((LIS_SOLVER)LIS_V2P(solver),buf);

	LIS_DEBUG_FUNC_OUT;
	return;
}

#endif

