/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * function                    | SOM |
 *-----------------------------+-----+
 * lis_matrix_set              | o   |
 * lis_matrix_setDLU           | o   |
 * lis_matrix_malloc           | o   |
 * lis_matrix_elements_copy    | o   |
 * lis_matrix_transpose        | o   |
 * lis_matrix_split            | o   |
 * lis_matrix_merge            | o   |
 *-----------------------------+-----+-----+
 * function                    |merge|split|
 *-----------------------------+-----+-----|
 * lis_matrix_convert          | o   |     |
 * lis_matrix_copy             | o   | o   |
 * lis_matrix_get_diagonal     | o   | o   |
 * lis_matrix_scaling          | o   | o   |
 * lis_matrix_scaling_symm     | o   | o   |
 * lis_matrix_normf            | o   | o   |
 * lis_matrix_sort             | o   | o   |
 * lis_matrix_solve            | xxx | o   |
 * lis_matrix_solvet           | xxx | o   |
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_dia"
int lis_matrix_set_dia(int nnd, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_SET);
	if( err ) return err;

	A->index       = index;
	A->value       = value;
	A->is_copy     = LIS_FALSE;
	A->status      = -LIS_MATRIX_DIA;
	A->nnd         = nnd;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_setDLU_dia"
int lis_matrix_setDLU_dia(int lnnd, int unnd, LIS_SCALAR *diag, int *lindex, LIS_SCALAR *lvalue,
						  int *uindex, LIS_SCALAR *uvalue, LIS_MATRIX A)
{
	int				i,err;
	LIS_MATRIX_DIAG	D;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_SET);
	if( err ) return err;

	A->L = lis_calloc(sizeof(struct LIS_MATRIX_CORE_STRUCT));
	if( A->L==NULL )
	{
		LIS_SETERR_MEM(sizeof(struct LIS_MATRIX_CORE_STRUCT));
		return LIS_OUT_OF_MEMORY;
	}
	A->U = lis_calloc(sizeof(struct LIS_MATRIX_CORE_STRUCT));
	if( A->U==NULL )
	{
		LIS_SETERR_MEM(sizeof(struct LIS_MATRIX_CORE_STRUCT));
		lis_matrix_DLU_destroy(A);
		return LIS_OUT_OF_MEMORY;
	}
	err = lis_matrix_diag_create(A->n,0,A->comm,&D);
	if( err )
	{
		lis_matrix_DLU_destroy(A);
		return err;
	}

	lis_free(D->value);
	D->value       = diag;
	A->D           = D;
	A->L->nnd      = lnnd;
	A->L->index    = lindex;
	A->L->value    = lvalue;
	A->U->nnd      = unnd;
	A->U->index    = uindex;
	A->U->value    = uvalue;
	A->is_copy     = LIS_FALSE;
	A->status      = -LIS_MATRIX_DIA;
	A->is_splited  = LIS_TRUE;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_malloc_dia"
int lis_matrix_malloc_dia(int n, int nnd, int **index, LIS_SCALAR **value)
{
	LIS_DEBUG_FUNC_IN;

	*index   = NULL;
	*value   = NULL;

	*index = (int *)lis_malloc( n*nnd*sizeof(int) );
	if( *index==NULL )
	{
		LIS_SETERR_MEM(n*nnd*sizeof(int));
		lis_free2(2,*index,*value);
		return LIS_OUT_OF_MEMORY;
	}
	*value = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( *value==NULL )
	{
		LIS_SETERR_MEM(n*nnd*sizeof(LIS_SCALAR));
		lis_free2(2,*index,*value);
		return LIS_OUT_OF_MEMORY;
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_matrix_elements_copy_dia"
int lis_matrix_elements_copy_dia(int n, int nnd, int *index, LIS_SCALAR *value,
								 int *o_index, LIS_SCALAR *o_value)
{
	int			i,j,is,ie;
	int			nprocs,my_rank;

	LIS_DEBUG_FUNC_IN;

	#ifdef _OPENMP
		nprocs = omp_get_max_threads();
	#else
		nprocs = 1;
	#endif

	memcpy(o_index,index,nnd*sizeof(int));
	#pragma omp parallel private(is,ie,my_rank)
	{
		#ifdef _OPENMP
			my_rank = omp_get_thread_num();
		#else
			my_rank = 0;
		#endif
		LIS_GET_ISIE(my_rank,nprocs,n,is,ie)

		memcpy(&o_value[is*nnd],&value[is*nnd],(ie-is)*nnd*sizeof(LIS_SCALAR));
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_copy_dia"
int lis_matrix_copy_dia(LIS_MATRIX Ain, LIS_MATRIX Aout)
{
	int			err;
	int			i,n,nnd,lnnd,unnd;
	int			*index;
	int			*lindex;
	int			*uindex;
	LIS_SCALAR	*value,*lvalue,*uvalue,*diag;

	LIS_DEBUG_FUNC_IN;

	n       = Ain->n;

	if( Ain->is_splited )
	{
		lnnd     = Ain->L->nnd;
		unnd     = Ain->U->nnd;
		lindex   = NULL;
		uindex   = NULL;
		diag     = NULL;

		err = lis_matrix_malloc_dia(n,lnnd,&lindex,&lvalue);
		if( err )
		{
			return err;
		}
		err = lis_matrix_malloc_dia(n,unnd,&uindex,&uvalue);
		if( err )
		{
			lis_free2(5,diag,uindex,lindex,uvalue,lvalue);
			return err;
		}
		diag = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR));
		if( diag==NULL )
		{
			lis_free2(5,diag,uindex,lindex,uvalue,lvalue);
			return err;
		}

		#pragma omp parallel for private(i)
		for(i=0;i<n;i++)
		{
			diag[i] = Ain->D->value[i];
		}
		lis_matrix_elements_copy_dia(n,lnnd,Ain->L->index,Ain->L->value,lindex,lvalue);
		lis_matrix_elements_copy_dia(n,unnd,Ain->U->index,Ain->U->value,uindex,uvalue);

		err = lis_matrix_setDLU_dia(lnnd,unnd,diag,lindex,lvalue,uindex,uvalue,Aout);
		if( err )
		{
			lis_free2(5,diag,uindex,lindex,uvalue,lvalue);
			return err;
		}
	}
	if( !Ain->is_splited || (Ain->is_splited && Ain->is_save) )
	{
		index   = NULL;
		value   = NULL;
		nnd     = Ain->nnd;
		err = lis_matrix_malloc_dia(n,nnd,&index,&value);
		if( err )
		{
			return err;
		}

		lis_matrix_elements_copy_dia(n,nnd,Ain->index,Ain->value,index,value);

		err = lis_matrix_set_dia(nnd,index,value,Aout);
		if( err )
		{
			lis_free2(2,index,value);
			return err;
		}
	}

	err = lis_matrix_assemble(Aout);
	if( err )
	{
		lis_matrix_storage_destroy(Aout);
		return err;
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_get_diagonal_dia"
int lis_matrix_get_diagonal_dia(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,is,ie,my_rank,nprocs;
	int n,nnd;

	LIS_DEBUG_FUNC_IN;

	n    = A->n;
	nnd  = A->nnd;
	if( A->is_splited )
	{
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			d[i] = A->D->value[i];
		}
	}
	else
	{
		#ifdef _OPENMP
			nprocs  = omp_get_max_threads();
			for(j=0;j<nnd;j++)
			{
				if( A->index[j]==0 ) break;
			}
			#pragma omp parallel private(is,ie,my_rank)
			{
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
				memcpy(&d[is],&A->value[is*nnd+j*(ie-is)],(ie-is)*sizeof(LIS_SCALAR));
			}
		#else
			for(j=0;j<nnd;j++)
			{
				if( A->index[j]==0 ) break;
			}
			for(i=0;i<n;i++)
			{
				d[i] = A->value[j*n+i];
			}
		#endif
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_scaling_dia"
int lis_matrix_scaling_dia(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,is,ie,js,je,jj,ii;
	int n,np,nnd,k;
	int my_rank,nprocs;


	LIS_DEBUG_FUNC_IN;

	n      = A->n;
	np     = A->np;
	if( A->is_splited )
	{
		#ifdef _OPENMP
			#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
			{
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie)
				for(i=is;i<ie;i++)
				{
					A->D->value[i] = 1.0;
				}
				for(j=0;j<A->L->nnd;j++)
				{
					jj = A->L->index[j];
					js = _max(is,-jj);
					#ifdef USE_MPI
						je = jj<=(np-n)?ie:_min(ie,np-jj);
					#else
						je = _min(ie,n-jj);
					#endif
					k  = is*A->L->nnd + j*(ie-is);
					ii = js-is;
					for(i=js;i<je;i++)
					{
						A->L->value[k + ii++] *= d[i];
					}
				}
				for(j=0;j<A->U->nnd;j++)
				{
					jj = A->U->index[j];
					js = _max(is,-jj);
					#ifdef USE_MPI
						je = jj<=(np-n)?ie:_min(ie,np-jj);
					#else
						je = _min(ie,n-jj);
					#endif
					k  = is*A->U->nnd + j*(ie-is);
					ii = js-is;
					for(i=js;i<je;i++)
					{
						A->U->value[k + ii++] *= d[i];
					}
				}
			}
		#else
			for(i=0;i<n;i++)
			{
				A->D->value[i] = 1.0;
			}
			for(j=0;j<A->L->nnd;j++)
			{
				jj = A->L->index[j];
				js = _max(0,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(n,np-jj);
				#else
					je = _min(n,n-jj);
				#endif
				for(i=js;i<je;i++)
				{
					A->L->value[j*n + i] *= d[i];
				}
			}
			for(j=0;j<A->U->nnd;j++)
			{
				jj = A->U->index[j];
				js = _max(0,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(n,np-jj);
				#else
					je = _min(n,n-jj);
				#endif
				for(i=js;i<je;i++)
				{
					A->U->value[j*n + i] *= d[i];
				}
			}
		#endif
	}
	else
	{
		nnd    = A->nnd;
		#ifdef _OPENMP
			#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
			{
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie)
				for(j=0;j<nnd;j++)
				{
					jj = A->index[j];
					js = _max(is,-jj);
					#ifdef USE_MPI
						je = jj<=(np-n)?ie:_min(ie,np-jj);
					#else
						je = _min(ie,n-jj);
					#endif
					k  = is*nnd + j*(ie-is);
					ii = js-is;
					for(i=js;i<je;i++)
					{
						A->value[k + ii] *= d[i];
						ii++;
					}
				}
			}
		#else
			for(j=0;j<nnd;j++)
			{
				jj = A->index[j];
				js = _max(0,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(n,np-jj);
				#else
					je = _min(n,n-jj);
				#endif
				for(i=js;i<je;i++)
				{
					A->value[j*n + i] *= d[i];
				}
			}
		#endif
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_scaling_symm_dia"
int lis_matrix_scaling_symm_dia(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,is,ie,js,je,jj,ii;
	int n,np,nnd,k;
	int my_rank,nprocs;

	LIS_DEBUG_FUNC_IN;

	n      = A->n;
	np     = A->np;
	nnd  = A->nnd;
	if( A->is_splited )
	{
		#ifdef _OPENMP
			#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
			{
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie)
				for(i=is;i<ie;i++)
				{
					A->D->value[i] = 1.0;
				}
				for(j=0;j<A->L->nnd;j++)
				{
					jj = A->L->index[j];
					js = _max(is,-jj);
					#ifdef USE_MPI
						je = jj<=(np-n)?ie:_min(ie,np-jj);
					#else
						je = _min(ie,n-jj);
					#endif
					k  = is*A->L->nnd + j*(ie-is);
					ii = js-is;
					for(i=js;i<je;i++)
					{
						A->L->value[k + ii++] *= d[i]*d[i+A->L->index[j]];;
					}
				}
				for(j=0;j<A->U->nnd;j++)
				{
					jj = A->U->index[j];
					js = _max(is,-jj);
					#ifdef USE_MPI
						je = jj<=(np-n)?ie:_min(ie,np-jj);
					#else
						je = _min(ie,n-jj);
					#endif
					k  = is*A->U->nnd + j*(ie-is);
					ii = js-is;
					for(i=js;i<je;i++)
					{
						A->U->value[k + ii++] *= d[i]*d[i+A->U->index[j]];;
					}
				}
			}
		#else
			for(i=0;i<n;i++)
			{
				A->D->value[i] = 1.0;
			}
			for(j=0;j<A->L->nnd;j++)
			{
				jj = A->L->index[j];
				js = _max(0,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(n,np-jj);
				#else
					je = _min(n,n-jj);
				#endif
				for(i=js;i<je;i++)
				{
					A->L->value[j*n + i] *= d[i]*d[i+A->L->index[j]];;
				}
			}
			for(j=0;j<A->U->nnd;j++)
			{
				jj = A->U->index[j];
				js = _max(0,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(n,np-jj);
				#else
					je = _min(n,n-jj);
				#endif
				for(i=js;i<je;i++)
				{
					A->U->value[j*n + i] *= d[i]*d[i+A->U->index[j]];;
				}
			}
		#endif
	}
	else
	{
		nnd    = A->nnd;
		#ifdef _OPENMP
			#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
			{
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie)
				for(j=0;j<nnd;j++)
				{
					jj = A->index[j];
					js = _max(is,-jj);
					#ifdef USE_MPI
						je = jj<=(np-n)?ie:_min(ie,np-jj);
					#else
						je = _min(ie,n-jj);
					#endif
					k  = is*nnd + j*(ie-is);
					ii = js-is;
					for(i=js;i<je;i++)
					{
						A->value[k + ii] *= d[i]*d[i+A->index[j]];
						ii++;
					}
				}
			}
		#else
			for(j=0;j<nnd;j++)
			{
				jj = A->index[j];
				js = _max(0,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(n,np-jj);
				#else
					je = _min(n,n-jj);
				#endif
				for(i=js;i<je;i++)
				{
					A->value[j*n + i] *= d[i]*d[i+A->index[j]];
				}
			}
		#endif
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_normf_dia"
int lis_matrix_normf_dia(LIS_MATRIX A, LIS_SCALAR *nrm)
{
	int i,j;
	int n;
	LIS_SCALAR sum;

	LIS_DEBUG_FUNC_IN;

	n    = A->n;
	sum  = (LIS_SCALAR)0;
	if( A->is_splited )
	{
		#pragma omp parallel for reduction(+:sum) private(i,j)
		for(i=0; i<n; i++)
		{
			sum += A->D->value[i]*A->D->value[i];
			for(j=A->L->index[i];j<A->L->index[i+1];j++)
			{
				sum += A->L->value[j]*A->L->value[j];
			}
			for(j=A->U->index[i];j<A->U->index[i+1];j++)
			{
				sum += A->U->value[j]*A->U->value[j];
			}
		}
	}
	else
	{
		#pragma omp parallel for reduction(+:sum) private(i,j)
		for(i=0; i<n; i++)
		{
			sum += A->value[i]*A->value[i];
			for(j=A->index[i];j<A->index[i+1];j++)
			{
				sum += A->value[j]*A->value[j];
			}
		}
	}
	*nrm = sqrt(sum);
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_transpose_dia"
int lis_matrix_transpose_dia(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
	int			err;

	LIS_DEBUG_FUNC_IN;

/*	err = lis_matrix_convert_dia2ccs(Ain,Aout);*/
	(*Aout)->matrix_type = LIS_MATRIX_DIA;
	(*Aout)->status      = LIS_MATRIX_DIA;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_split_dia"
int lis_matrix_split_dia(LIS_MATRIX A)
{
	int				i,j,n,kl,ku,nnd;
	int				lnnd,unnd;
	int				nprocs,my_rank,is,ie;
	int				err;
	int				*lindex,*uindex;
	LIS_SCALAR		*lvalue,*uvalue;
	LIS_MATRIX_DIAG	D;

	LIS_DEBUG_FUNC_IN;

	n        = A->n;
	nnd      = A->nnd;
	lnnd     = 0;
	unnd     = 0;
	D        = NULL;
	lindex   = NULL;
	lvalue   = NULL;
	uindex   = NULL;
	uvalue   = NULL;

	for(j=0;j<nnd;j++)
	{
		if( A->index[j]<0 )
		{
			lnnd++;
		}
		else if( A->index[j]>0 )
		{
			unnd++;
		}
	}

	err = lis_matrix_LU_create(A);
	if( err )
	{
		return err;
	}
	err = lis_matrix_malloc_dia(n,lnnd,&lindex,&lvalue);
	if( err )
	{
		return err;
	}
	err = lis_matrix_malloc_dia(n,unnd,&uindex,&uvalue);
	if( err )
	{
		lis_free2(4,lindex,lvalue,uindex,uvalue);
		return err;
	}
	err = lis_matrix_diag_duplicateM(A,&D);
	if( err )
	{
		lis_free2(4,lindex,lvalue,uindex,uvalue);
		return err;
	}

	#ifdef _OPENMP
		kl = 0;
		ku = 0;
		nprocs  = omp_get_max_threads();
		for(j=0;j<nnd;j++)
		{
			if( A->index[j]<0 )
			{
				lindex[kl]   = A->index[j];
				#pragma omp parallel private(i,is,ie,my_rank)
				{
					my_rank = omp_get_thread_num();
					LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
					memcpy(&lvalue[is*lnnd+kl*(ie-is)],&A->value[is*nnd+j*(ie-is)],(ie-is)*sizeof(LIS_SCALAR));
				}
				kl++;
			}
			else if( A->index[j]>0 )
			{
				uindex[ku]   = A->index[j];
				#pragma omp parallel private(i,is,ie,my_rank)
				{
					my_rank = omp_get_thread_num();
					LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
					memcpy(&uvalue[is*unnd+ku*(ie-is)],&A->value[is*nnd+j*(ie-is)],(ie-is)*sizeof(LIS_SCALAR));
				}
				ku++;
			}
			else
			{
				#pragma omp parallel private(i,is,ie,my_rank)
				{
					my_rank = omp_get_thread_num();
					LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
					for(i=is;i<ie;i++)
					{
						D->value[i] = A->value[is*nnd+j*(ie-is)+i-is];
					}
				}
			}
		}
	#else
		lnnd = 0;
		unnd = 0;
		for(j=0;j<nnd;j++)
		{
			if( A->index[j]<0 )
			{
				lindex[lnnd]   = A->index[j];
				for(i=0;i<n;i++)
				{
					lvalue[lnnd*n+i] = A->value[j*n+i];
				}
				lnnd++;
			}
			else if( A->index[j]>0 )
			{
				uindex[unnd]   = A->index[j];
				for(i=0;i<n;i++)
				{
					uvalue[unnd*n+i] = A->value[j*n+i];
				}
				unnd++;
			}
			else
			{
				for(i=0;i<n;i++)
				{
					D->value[i] = A->value[j*n+i];
				}
			}
		}
	#endif
	A->L->nnd     = lnnd;
	A->L->index   = lindex;
	A->L->value   = lvalue;
	A->U->nnd     = unnd;
	A->U->index   = uindex;
	A->U->value   = uvalue;
	A->D          = D;
	A->is_splited = LIS_TRUE;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_matrix_merge_dia"
int lis_matrix_merge_dia(LIS_MATRIX A)
{
	int				i,j,k,n;
	int				nprocs,my_rank,is,ie;
	int				nnd;
	int				err;
	int				*index;
	LIS_SCALAR		*value;

	LIS_DEBUG_FUNC_IN;


	n       = A->n;
	is      = A->is;
	index   = NULL;
	value   = NULL;
	nnd     = A->L->nnd + A->U->nnd + 1;

	err = lis_matrix_malloc_dia(n,nnd,&index,&value);
	if( err )
	{
		return err;
	}

	#ifdef _OPENMP
		nprocs  = omp_get_max_threads();
		k = 0;
		for(j=0;j<A->L->nnd;j++)
		{
			index[k]   = A->L->index[j];
			#pragma omp parallel private(i,is,ie,my_rank)
			{
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
				memcpy(&value[is*nnd+k*(ie-is)],&A->L->value[is*A->L->nnd+j*(ie-is)],(ie-is)*sizeof(LIS_SCALAR));
			}
			k++;
		}
		index[k]   = 0;
		#pragma omp parallel private(i,is,ie,my_rank)
		{
			my_rank = omp_get_thread_num();
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
			memcpy(&value[is*nnd+k*(ie-is)],&A->D->value[is],(ie-is)*sizeof(LIS_SCALAR));
		}
		k++;
		for(j=0;j<A->U->nnd;j++)
		{
			index[k]   = A->U->index[j];
			#pragma omp parallel private(i,is,ie,my_rank)
			{
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
				memcpy(&value[is*nnd+k*(ie-is)],&A->U->value[is*A->U->nnd+j*(ie-is)],(ie-is)*sizeof(LIS_SCALAR));
			}
			k++;
		}
	#else
		k = 0;
		for(j=0;j<A->L->nnd;j++)
		{
			index[k]   = A->L->index[j];
			for(i=0;i<n;i++)
			{
				value[k*n+i] = A->L->value[j*n+i];
			}
			k++;
		}
		index[k]   = 0;
		for(i=0;i<n;i++)
		{
			value[k*n+i] = A->D->value[i];
		}
		k++;
		for(j=0;j<A->U->nnd;j++)
		{
			index[k]   = A->U->index[j];
			for(i=0;i<n;i++)
			{
				value[k*n+i] = A->U->value[j*n+i];
			}
			k++;
		}
	#endif

	A->nnd        = nnd;
	A->value      = value;
	A->index      = index;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_sort_dia"
int lis_matrix_sort_dia(LIS_MATRIX A)
{
	int i,j,n;

	LIS_DEBUG_FUNC_IN;

	if( !A->is_sorted )
	{
		n = A->n;
		if( A->is_splited )
		{
			#pragma omp parallel for private(i)
			for(i=0;i<n;i++)
			{
				lis_sort_id(A->L->ptr[i],A->L->ptr[i+1]-1,A->L->index,A->L->value);
				lis_sort_id(A->U->ptr[i],A->U->ptr[i+1]-1,A->U->index,A->U->value);
			}
		}
		else
		{
			#pragma omp parallel for private(i)
			for(i=0;i<n;i++)
			{
				lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
			}
		}
		A->is_sorted = LIS_TRUE;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_solve_dia"
int lis_matrix_solve_dia(LIS_MATRIX A, LIS_VECTOR B, LIS_VECTOR X, int flag)
{
	int i,j,n;
	LIS_SCALAR	t;
	LIS_SCALAR	*b,*x;

	LIS_DEBUG_FUNC_IN;

	n  = A->n;
	b  = B->value;
	x  = X->value;

	switch(flag)
	{
	case LIS_MATRIX_LOWER:
		for(i=0;i<n;i++)
		{
			t = b[i];
			for(j=0;j<A->L->nnd;j++)
			{
				if( i+A->L->index[j] >= 0 ) t -= A->L->value[j*n + i] * x[i + A->L->index[j]];
			}
			x[i]   = t * A->WD->value[i];
		}
		break;
	case LIS_MATRIX_UPPER:
		for(i=n-1;i>=0;i--)
		{
			t = b[i];
			for(j=0;j<A->U->nnd;j++)
			{
				if( i+A->U->index[j] < n ) t -= A->U->value[j*n + i] * x[i + A->U->index[j]];
			}
			x[i]   = t * A->WD->value[i];
		}
		break;
	case LIS_MATRIX_SSOR:
		for(i=0;i<n;i++)
		{
			t = b[i];
			for(j=0;j<A->L->nnd;j++)
			{
				if( i+A->L->index[j] >= 0 ) t -= A->L->value[j*n + i] * x[i + A->L->index[j]];
			}
			x[i]   = t * A->WD->value[i];
		}
		for(i=n-1;i>=0;i--)
		{
			t = 0.0;
			for(j=0;j<A->U->nnd;j++)
			{
				if( i+A->U->index[j] < n ) t += A->U->value[j*n + i] * x[i + A->U->index[j]];
			}
			x[i]  -= t * A->WD->value[i];
		}
		break;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_solvet_dia"
int lis_matrix_solvet_dia(LIS_MATRIX A, LIS_VECTOR B, LIS_VECTOR X, int flag)
{
	int i,j,n;
	LIS_SCALAR	t;
	LIS_SCALAR	*b,*x;

	LIS_DEBUG_FUNC_IN;

	n  = A->n;
	b  = B->value;
	x  = X->value;

	lis_vector_copy(B,X);
	switch(flag)
	{
	case LIS_MATRIX_LOWER:
		for(i=0;i<n;i++)
		{
			x[i]  = x[i] * A->WD->value[i];
			for(j=0;j<A->U->nnd;j++)
			{
				if( i+A->U->index[j] < n ) x[i + A->U->index[j]] -= A->U->value[j*n + i] * x[i];
			}
		}
		break;
	case LIS_MATRIX_UPPER:
		for(i=n-1;i>=0;i--)
		{
			x[i]  = x[i] * A->WD->value[i];
			for(j=0;j<A->L->nnd;j++)
			{
				if( i+A->L->index[j] >= 0 ) x[i + A->L->index[j]] -= A->L->value[j*n + i] * x[i];
			}
		}
		break;
	case LIS_MATRIX_SSOR:
		for(i=0;i<n;i++)
		{
			t  = x[i] * A->WD->value[i];
			for(j=0;j<A->U->nnd;j++)
			{
				if( i+A->U->index[j] < n ) x[i + A->U->index[j]] -= A->U->value[j*n + i] * t;
			}
		}
		for(i=n-1;i>=0;i--)
		{
			x[i]  = x[i] * A->WD->value[i];
			t     = x[i];
			for(j=0;j<A->L->nnd;j++)
			{
				if( i+A->L->index[j] >= 0 ) x[i + A->L->index[j]] -= A->L->value[j*n + i] * t;
			}
		}
		break;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_convert_crs2dia"
int lis_matrix_convert_crs2dia(LIS_MATRIX Ain, LIS_MATRIX Aout)
{
	int			i,j,jj,k;
	int			err;
	int			n,nnz,nnd,nprocs,my_rank;
	int			is,ie;
	int			*iw;
	int			*index;
	LIS_SCALAR	*value;

	LIS_DEBUG_FUNC_IN;

	n       = Ain->n;
	nnz		= Ain->nnz;

	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( nnz*sizeof(int) );
	if( iw==NULL )
	{
		LIS_SETERR_MEM(nnz*sizeof(int));
		return LIS_ERR_OUT_OF_MEMORY;
	}

	lis_matrix_sort_crs(Ain);

	#pragma omp parallel for private(i,j)
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			iw[j] = Ain->index[j] - i;
		}
	}
	lis_sort_i(0,nnz-1,iw);
	nnd = 1;
	k   = iw[0];
	for(i=1;i<nnz;i++)
	{
		if( k!=iw[i] )
		{
			k = iw[i];
			nnd++;
		}
	}

	err = lis_matrix_malloc_dia(n,nnd,&index,&value);
	if( err )
	{
		lis_free(iw);
		return err;
	}

	/* convert dia */
	k        = iw[0];
	index[0] = k;
	j        = 1;
	for(i=1;i<nnz;i++)
	{
		if( k!=iw[i] )
		{
			k = iw[i];
			index[j] = k;
			j++;
		}
	}
	#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
	{
		#ifdef _OPENMP
			my_rank = omp_get_thread_num();
			nprocs  = omp_get_max_threads();
		#else
			my_rank = 0;
			nprocs  = 1;
		#endif
		LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
		memset(&value[is*nnd],0,(ie-is)*nnd*sizeof(LIS_SCALAR));

		for(i=is;i<ie;i++)
		{
			k = 0;
			for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
			{
				jj=Ain->index[j] - i;
				while( jj!=index[k] ) k++;
				value[is*nnd + k*(ie-is) + i-is] = Ain->value[j];
			}
		}
	}

	err = lis_matrix_set_dia(nnd,index,value,Aout);
	if( err )
	{
		lis_free2(3,index,value,iw);
		return err;
	}
	err = lis_matrix_assemble(Aout);
	if( err )
	{
		lis_free(iw);
		lis_matrix_storage_destroy(Aout);
		return err;
	}
	lis_free(iw);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_convert_dia2crs"
int lis_matrix_convert_dia2crs(LIS_MATRIX Ain, LIS_MATRIX Aout)
{
	int			i,j,jj,k,l;
	int			err,js,je;
	int			n,np,nnz,nnd,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;

	LIS_DEBUG_FUNC_IN;

	n       = Ain->n;
	np      = Ain->np;
	nnd     = Ain->nnd;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( iw==NULL )
	{
		LIS_SETERR_MEM((n+1)*sizeof(int));
		return LIS_ERR_OUT_OF_MEMORY;
	}

	iw[0]  = 0;
	#pragma omp parallel private(i,j,k,is,ie,js,je,jj,my_rank)
	{
		#ifdef _OPENMP
			nprocs  = omp_get_max_threads();
			my_rank = omp_get_thread_num();
		#else
			my_rank = 0;
			nprocs  = 1;
		#endif
		LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
		memset(&iw[is+1],0,(ie-is)*sizeof(int));
		k = ie-is;

		for(j=0;j<nnd;j++)
		{
			jj = Ain->index[j];
			js = _max(is,-jj) -is;
			#ifdef USE_MPI
				je = jj<=(np-n)?ie:_min(ie,np-jj);
			#else
				je = _min(ie,n-jj);
			#endif
			je -= is;
			for(i=js;i<je;i++)
			{ 
				if( Ain->value[is*nnd + j*k + i]!=(LIS_SCALAR)0.0 )
				{
					iw[i+is+1]++;
				}
			}
		}
	}
	for(i=0;i<n;i++)
	{
		iw[i+1] += iw[i];
	}
	nnz = iw[n];

	err = lis_matrix_malloc_crs(n,nnz,&ptr,&index,&value);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		return err;
	}

	/* convert crs */
	ptr[0] = 0;
	#pragma omp parallel private(i,j,k,l,is,ie,js,je,jj,my_rank)
	{
		#ifdef _OPENMP
			nprocs  = omp_get_max_threads();
			my_rank = omp_get_thread_num();
		#else
			my_rank = 0;
			nprocs  = 1;
		#endif
		LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
		l = ie-is;

		for(i=is;i<ie;i++)
		{
			ptr[i+1] = iw[i+1];
		}
		for(j=0;j<nnd;j++)
		{
			jj = Ain->index[j];
			js = _max(is,-jj) - is;
			#ifdef USE_MPI
				je = jj<=(np-n)?ie:_min(ie,np-jj);
			#else
				je = _min(ie,n-jj);
			#endif
			je -= is;
			for(i=js;i<je;i++)
			{ 
				if( Ain->value[is*nnd + j*l + i]!=(LIS_SCALAR)0.0 )
				{
					k        = iw[i+is]++;
					value[k] = Ain->value[is*nnd + j*l + i];
					index[k] = i+is + jj;
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		return err;
	}
	err = lis_matrix_assemble(Aout);
	if( err )
	{
		lis_free(iw);
		lis_matrix_storage_destroy(Aout);
		return err;
	}
	lis_free(iw);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

