/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

#undef __FUNC__
#define __FUNC__ "lis_matvec"
int lis_matvec(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	LIS_SCALAR *x,*y;

	LIS_DEBUG_FUNC_IN;

	if( X->precision==LIS_PRECISION_DEFAULT )
	{
		x = X->value;
		y = Y->value;

		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_crs_xxx[0][0](A, x, y);
			break;
		case LIS_MATRIX_BSR:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			if( A->bnr<=4 && A->bnc<=4 )
			{
				lis_matvec_bsr_xxx[A->bnr-1][A->bnc-1](A, x, y);
			}
			else
			{
				lis_matvec_bsr(A, x, y);
			}
			break;
		case LIS_MATRIX_CCS:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_ccs(A, x, y);
			break;
		case LIS_MATRIX_BSC:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_bsc(A, x, y);
			break;
		case LIS_MATRIX_MSR:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_msr(A, x, y);
			break;
		case LIS_MATRIX_ELL:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_ell(A, x, y);
			break;
		case LIS_MATRIX_DIA:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_dia(A, x, y);
			break;
		case LIS_MATRIX_JDS:
			#ifdef USE_MPI
				#ifndef USE_OVERLAP
					LIS_MATVEC_SENDRECV;
				#else
					LIS_MATVEC_REALLOC;
				#endif
			#endif
			lis_matvec_jds(A, x, y);
			break;
		case LIS_MATRIX_VBR:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_vbr(A, x, y);
			break;
		case LIS_MATRIX_DNS:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_dns(A, x, y);
			break;
		case LIS_MATRIX_COO:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			lis_matvec_coo(A, x, y);
			break;
	/*
		case LIS_MATRIX_SDIA:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_sdia(A, x, y);
			break;
		case LIS_MATRIX_LJDS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_ljds(A, x, y);
			break;
		case LIS_MATRIX_PCRS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_pcrs(A, x, y);
			break;
		case LIS_MATRIX_LCRS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_lcrs(A, x, y);
			break;
		case LIS_MATRIX_CJD:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_cjd(A, x, y);
			break;
		case LIS_MATRIX_TJD:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_tjd(A, x, y);
			break;
		case LIS_MATRIX_BJD:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_bjd(A, x, y);
			break;
		case LIS_MATRIX_BCR:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_bcr(A, x, y);
			break;
		case LIS_MATRIX_LBSR:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_lbsr(A, x, y);
			break;
	*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}
#ifdef USE_QUAD_PRECISION
	else
	{

		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			#ifdef USE_MPI
				lis_send_recv_mp(A->commtable,X);
			#endif
			#ifndef USE_FMA2_SSE2
				lis_matvec_crs_xxxP[0][0](A, X, Y);
			#else
				lis_matvec_crs_xxxP[0][1](A, X, Y);
			#endif
			break;
		case LIS_MATRIX_CCS:
			#ifdef USE_MPI
				lis_send_recv_mp(A->commtable,X);
			#endif
			#ifndef USE_FMA2_SSE2
				lis_matvec_ccs_mp(A, X, Y);
			#else
				lis_matvec_ccs_mp2(A, X, Y);
			#endif
			break;
	/*
		case LIS_MATRIX_BSR:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			if( A->bnr<=4 && A->bnc<=4 )
			{
				lis_matvec_bsr_xxx[A->bnr-1][A->bnc-1](A, x, y);
			}
			else
			{
				lis_matvec_bsr(A, x, y);
			}
			break;
		case LIS_MATRIX_CCS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_crs(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_BSC:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_bsr(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_MSR:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_msr(A, x, y);
			break;
		case LIS_MATRIX_DIA:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_dia(A, x, y);
			break;
		case LIS_MATRIX_SDIA:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_sdia(A, x, y);
			break;
		case LIS_MATRIX_ELL:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_ell(A, x, y);
			break;
		case LIS_MATRIX_JDS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_jds(A, x, y);
			break;
		case LIS_MATRIX_LJDS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_ljds(A, x, y);
			break;
		case LIS_MATRIX_PCRS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_pcrs(A, x, y);
			break;
		case LIS_MATRIX_LCRS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_lcrs(A, x, y);
			break;
		case LIS_MATRIX_CJD:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_cjd(A, x, y);
			break;
		case LIS_MATRIX_TJD:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_tjd(A, x, y);
			break;
		case LIS_MATRIX_BJD:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_bjd(A, x, y);
			break;
		case LIS_MATRIX_BCR:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_bcr(A, x, y);
			break;
		case LIS_MATRIX_LBSR:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_lbsr(A, x, y);
			break;
		case LIS_MATRIX_VBR:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_vbr(A, x, y);
			break;
		case LIS_MATRIX_DNS:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_dns(A, x, y);
			break;
		case LIS_MATRIX_COO:
			#ifdef USE_MPI
			if( A->np>X->np )
			{
				X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
				if( X->value==NULL )
				{
					SETERR("out of memory\n");
					return;
				}
				x = X->value;
			}
			lis_send_recv(A->commtable,x);
			#endif
			lis_matvec_coo(A, x, y);
			break;
	*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}
#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matvect"
int lis_matvect(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	LIS_SCALAR *x,*y;

	LIS_DEBUG_FUNC_IN;

	x = X->value;
	y = Y->value;

	if( X->precision==LIS_PRECISION_DEFAULT )
	{
		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_crs(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_BSR:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_bsr(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_CCS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_ccs(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_BSC:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
#if 0
			if( A->bnr<=4 && A->bnc<=4 )
			{
				lis_matvec_bsr_xxx[A->bnr-1][A->bnc-1](A, x, y);
			}
			else
			{
				lis_matvec_bsr(A, x, y);
			}
#endif
			lis_matvect_bsc(A, x, y);
			break;
		case LIS_MATRIX_MSR:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_msr(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_ELL:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_ell(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_JDS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_jds(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_DIA:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_dia(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_VBR:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_vbr(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_DNS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_dns(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_COO:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_coo(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
	/*
		case LIS_MATRIX_DNS:
			lis_matvect_dns(A, x, y);
			break;
	*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}
#ifdef USE_QUAD_PRECISION
	else
	{
		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			#ifndef USE_FMA2_SSE2
				lis_matvect_crs_mp(A, X, Y);
			#else
				lis_matvect_crs_mp2(A, X, Y);
			#endif
			#ifdef USE_MPI
				lis_reduce_mp(A->commtable,Y);
			#endif
			break;
		case LIS_MATRIX_CCS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			#ifndef USE_FMA2_SSE2
				lis_matvect_ccs_u1_1P(A, X, Y);
			#else
				lis_matvect_ccs_u1_2P(A, X, Y);
			#endif
			#ifdef USE_MPI
				lis_reduce_mp(A->commtable,Y);
			#endif
			break;
			/*
		case LIS_MATRIX_BSR:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_bsr(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_CCS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_ccs(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_BSC:
			#ifdef USE_MPI
				LIS_MATVEC_SENDRECV;
			#endif
			if( A->bnr<=4 && A->bnc<=4 )
			{
				lis_matvec_bsr_xxx[A->bnr-1][A->bnc-1](A, x, y);
			}
			else
			{
				lis_matvec_bsr(A, x, y);
			}
			break;
		case LIS_MATRIX_MSR:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_msr(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_ELL:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_ell(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_JDS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_jds(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_DIA:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_dia(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_DNS:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_dns(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
		case LIS_MATRIX_COO:
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE0;
			#endif
			lis_matvect_coo(A, x, y);
			#ifdef USE_MPI
				LIS_MATVEC_REDUCE;
			#endif
			break;
			*/
	/*
		case LIS_MATRIX_VBR:
			lis_matvect_vbr(A, x, y);
			break;
		case LIS_MATRIX_DNS:
			lis_matvect_dns(A, x, y);
			break;
	*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}
#endif
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
