/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef USE_SSE2
	#include <emmintrin.h>
#endif
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

void lis_matvec_ccs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n,np,nprocs;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	np   = A->np;
	if( A->is_splited )
	{
		for(i=0; i<n; i++)
		{
			y[i] = A->D->value[i]*x[i];
		}
		for(i=0; i<np; i++)
		{
			js = A->L->ptr[i];
			je = A->L->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->L->index[j];
				y[jj] += A->L->value[j] * t;
			}
			js = A->U->ptr[i];
			je = A->U->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->U->index[j];
				y[jj] += A->U->value[j] * t;
			}
		}
	}
	else
	{
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<np; i++)
				{
					js = A->ptr[i];
					je = A->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<n;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
		#else
			for(i=0; i<n; i++)
			{
				y[i] = 0.0;
			}
			for(i=0; i<np; i++)
			{
				js = A->ptr[i];
				je = A->ptr[i+1];
				t = x[i];
				for(j=js;j<je;j++)
				{
					jj  = A->index[j];
					y[jj] += A->value[j] * t;
				}
			}
		#endif
	}
}

void lis_matvect_ccs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n,np;
	LIS_SCALAR t;

	n    = A->n;
	np   = A->np;
	if( A->is_splited )
	{
		for(i=0; i<np; i++)
		{
			js = A->L->ptr[i];
			je = A->L->ptr[i+1];
			t =  A->D->value[i]*x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->L->index[j];
				t += A->L->value[j] * x[jj];
			}
			js = A->U->ptr[i];
			je = A->U->ptr[i+1];
			for(j=js;j<je;j++)
			{
				jj  = A->U->index[j];
				t += A->U->value[j] * x[jj];
			}
			y[i] = t;
		}
	}
	else
	{
		#pragma omp parallel for private(i,j,js,je,jj,t)
		for(i=0; i<np; i++)
		{
			js = A->ptr[i];
			je = A->ptr[i+1];
			t = 0.0;
			for(j=js;j<je;j++)
			{
				jj  = A->index[j];
				t += A->value[j] * x[jj];
			}
			y[i] = t;
		}
	}
}

#ifdef USE_QUAD_PRECISION
void lis_matvec_ccs_mp(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	int i,j,k,js,je,jj;
	int n,np,nprocs;
	LIS_SCALAR_M tt0;
	LIS_SCALAR		*x,*y,*xl,*yl;
	LIS_SCALAR *w;
	LIS_QUAD_DECLAR;

	n    = A->n;
	np   = A->np;
	x     = X->value;
	y     = Y->value;
	xl    = X->value_lo;
	yl    = Y->value_lo;
	tt0.hi = &X->work[0];
	tt0.lo = &X->work[1];
	if( A->is_splited )
	{
		/*
		for(i=0; i<np; i++)
		{
			y[i] = A->D->value[i] * x[i];
		}
		for(i=0; i<n; i++)
		{
			js = A->L->ptr[i];
			je = A->L->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->L->index[j];
				y[jj] += A->L->value[j] * t;
			}
		}
		for(i=0; i<n; i++)
		{
			js = A->U->ptr[i];
			je = A->U->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->U->index[j];
				y[jj] += A->U->value[j] * t;
			}
		}
		*/
	}
	else
	{
		#ifdef _OPENMP
			/*
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<np; i++)
				{
					js = A->ptr[i];
					je = A->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<n;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
			*/
		#else
			for(i=0; i<n; i++)
			{
				y[i]  = 0.0;
				yl[i] = 0.0;
			}
			for(i=0; i<np; i++)
			{
				js = A->ptr[i];
				je = A->ptr[i+1];
				tt0.hi[0] = x[i];
				tt0.lo[0] = xl[i];
				for(j=js;j<je;j++)
				{
					jj  = A->index[j];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(y[jj],yl[jj],y[jj],yl[jj],tt0.hi[0],tt0.lo[0],A->value[j]);
					#else
						LIS_QUAD_FMAD_SSE2(y[jj],yl[jj],y[jj],yl[jj],tt0.hi[0],tt0.lo[0],A->value[j]);
					#endif
				}
			}
		#endif
	}
}

void lis_matvec_ccs_mp2(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	int i,j,k,js,je,j0,j1;
	int n,np,nprocs;
	LIS_SCALAR_M tt0,yy0;
	LIS_SCALAR		*x,*y,*xl,*yl;
	LIS_SCALAR *w;
	LIS_QUAD_DECLAR;

	n    = A->n;
	np   = A->np;
	x     = X->value;
	y     = Y->value;
	xl    = X->value_lo;
	yl    = Y->value_lo;
	tt0.hi = &X->work[0];
	tt0.lo = &X->work[2];
	yy0.hi = &X->work[4];
	yy0.lo = &X->work[6];
	if( A->is_splited )
	{
	}
	else
	{
		#ifdef _OPENMP
		/*
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<np; i++)
				{
					js = A->ptr[i];
					je = A->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<n;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
			*/
		#else
			for(i=0; i<n; i++)
			{
				y[i]  = 0.0;
				yl[i] = 0.0;
			}
			for(i=0; i<np; i++)
			{
				js = A->ptr[i];
				je = A->ptr[i+1];
				tt0.hi[0] = x[i];
				tt0.hi[1] = x[i];
				tt0.lo[0] = xl[i];
				tt0.lo[1] = xl[i];
				for(j=js;j<je-1;j+=2)
				{
					j0  = A->index[j];
					j1  = A->index[j+1];
					#ifdef USE_SSE2
						LIS_QUAD_FMAD2_SSE2_ALIGN2(y[j0],yl[j0],y[j1],yl[j1],tt0.hi[0],tt0.lo[0],A->value[j]);
					#endif
				}
				for(;j<je;j++)
				{
					j0  = A->index[j];
					#ifdef USE_SSE2
						LIS_QUAD_FMAD_SSE2(y[j0],yl[j0],y[j0],yl[j0],tt0.hi[0],tt0.lo[0],A->value[j]);
					#endif
				}
			}
		#endif
	}
}

	void lis_matvect_ccs_u1_1P(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
	{
		int			i,j,k,h;
		int			n,np,nl,nz;
		int			is,ie,js,jj;
		int			is0;
		int			ie0;
		int			j0;
		int			*jj0;
		LIS_SCALAR			*vv0;
		LIS_SCALAR_M		tt0;
		LIS_MSCALAR		ax;
		LIS_SCALAR		*x,*y,*xl,*yl;
		LIS_QUAD_DECLAR;

		n     = A->n;
		np    = A->np;
		x     = X->value;
		y     = Y->value;
		xl    = X->value_lo;
		yl    = Y->value_lo;
		tt0.hi = &X->work[0];
		tt0.lo = &X->work[1];
		if( A->is_splited )
		{
			for(i=0;i<n;i++)
			{
/*				lisqp_mul_dd_d(&t0,&x[i],A->D->value[i]);*/

				is = A->L->ptr[i];
				ie = A->L->ptr[i+1];
				for(j=is;j<ie-0;j+=1)
				{
					j0  = A->L->index[j+0];
/*					lisqp_mul_dd_d(&ax,&x[j0],A->L->value[j+ 0]);
					lisqp_add(&t0,&t0,&ax);*/
				}
				is = A->U->ptr[i];
				ie = A->U->ptr[i+1];
				for(j=is;j<ie-0;j+=1)
				{
					j0  = A->U->index[j+0];
/*					lisqp_mul_dd_d(&ax,&x[j0],A->U->value[j+ 0]);
					lisqp_add(&t0,&t0,&ax);*/
				}
/*				y[i] = t0;*/
			}
		}
		else
		{
			jj0 = A->index;
			vv0 = A->value;
			for(i=0;i<np;i++)
			{
				tt0.hi[0] = tt0.lo[0] = 0.0;

				is = A->ptr[i];
				ie = A->ptr[i+1];
				for(j=is;j<ie-0;j+=1)
				{
					j0 = jj0[j+0];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(tt0.hi[0],tt0.lo[0],tt0.hi[0],tt0.lo[0],x[j0],xl[j0],vv0[j]);
					#else
						LIS_QUAD_FMAD_SSE2_ALIGN(tt0.hi[0],tt0.lo[0],tt0.hi[0],tt0.lo[0],x[j0],xl[j0],vv0[j]);
					#endif
				}
				y[i]  = tt0.hi[0];
				yl[i] = tt0.lo[0];
			}
		}
	}


void lis_matvect_ccs_u1_2P(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR_M	tt0;
	LIS_SCALAR_M	ax,xx,ax2,tt;
	LIS_SCALAR	*x,*y,*xl,*yl;
	LIS_QUAD_DECLAR;

	n     = A->n;
	np    = A->np;
	x     = X->value;
	y     = Y->value;
	xl    = X->value_lo;
	yl    = Y->value_lo;
	tt.hi = &X->work[0];
	tt.lo = &X->work[2];
	xx.hi = &X->work[4];
	xx.lo = &X->work[6];
	ax2.hi = &X->work[8];
	ax2.lo = &X->work[10];
	ax.hi = &X->work[12];
	ax.lo = &X->work[13];
	tt0.hi = &X->work[14];
	tt0.lo = &X->work[15];
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
/*			lisqp_mul_dd_d(tt0.hi,&x[2*i],A->D->value[i]);*/

		/*	lisqp_zero(&tt,2);*/
			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				xx.hi[0] = x[j0];
				xx.lo[0] = xl[j0];
				xx.hi[1] = x[j1];
				xx.lo[1] = xl[j1];
/*				lisqp_mul_dd_d2_inline(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->L->value[j+ 0]);
				lisqp_add2_inline(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d2(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->L->value[j+ 0]);
				lisqp_add2(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d(&ax,&x[j0],A->L->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);
				lisqp_mul_dd_d(&ax,&x[j1],A->L->value[j+ 1]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
/*			lisqp_add(tt0.hi,tt.hi,tt.hi);*/
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
/*				lisqp_mul_dd_d(ax.hi,&x[2*j0],A->L->value[j+ 0]);
				lisqp_add(tt0.hi,tt0.hi,ax.hi);
				lisqp_mul_dd_d(&ax,&x[j0],A->L->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				xx.hi[0] = x[j0];
				xx.lo[0] = xl[j0];
				xx.hi[1] = x[j1];
				xx.lo[1] = xl[j1];
/*				lisqp_mul_dd_d2_inline(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->U->value[j+ 0]);
				lisqp_add2_inline(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d2(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->U->value[j+ 0]);
				lisqp_add2(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d(&ax,&x[j0],A->U->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);
				lisqp_mul_dd_d(&ax,&x[j1],A->U->value[j+ 1]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
/*			lisqp_add(tt0.hi,tt.hi,tt.hi);*/
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
/*				lisqp_mul_dd_d(ax.hi,&x[2*j0],A->U->value[j+ 0]);
				lisqp_add(tt0.hi,tt0.hi,ax.hi);
				lisqp_mul_dd_d(&ax,&x[j0],A->U->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
			y[i]  = tt0.hi[0];
			yl[i] = tt0.lo[0];
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<np;i++)
		{
			tt0.hi[0] = tt0.lo[0] = 0.0;
			tt.hi[0] = tt.hi[1] = 0.0;
			tt.lo[0] = tt.lo[1] = 0.0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				#ifdef USE_SSE2
					LIS_QUAD_FMAD2_SSE2_ALIGN(tt.hi[0],tt.lo[0],tt.hi[0],tt.lo[0],x[j0],xl[j0],x[j1],xl[j1],vv0[j]);
				#endif
			}
			#ifdef USE_SSE2
				LIS_QUAD_ADD_SSE2(tt0.hi[0],tt0.lo[0],tt.hi[0],tt.lo[0],tt.hi[1],tt.lo[1]);
			#endif
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				#ifdef USE_SSE2
					LIS_QUAD_FMAD_SSE2_ALIGN(tt0.hi[0],tt0.lo[0],tt0.hi[0],tt0.lo[0],x[j0],xl[j0],vv0[j]);
				#endif
			}
			y[i]  = tt0.hi[0];
			yl[i] = tt0.lo[0];
			
		}
	}
}
#endif
