/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#ifdef USE_SSE2
	#include <emmintrin.h>
#endif
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

LIS_MATVEC_XXX lis_matvec_crs_xxx[1][8] = 
{
	{lis_matvec_crs_u1_1,lis_matvec_crs_u1_2,lis_matvec_crs_u1_3,lis_matvec_crs_u1_4,lis_matvec_crs_u1_5,lis_matvec_crs_u1_6,lis_matvec_crs_u1_7,lis_matvec_crs_u1_8,},
};

#ifdef USE_QUAD_PRECISION
	LIS_MATVEC_XXXP lis_matvec_crs_xxxP[1][8] = 
	{
		{lis_matvec_crs_u1_1P,lis_matvec_crs_u1_2P,NULL,NULL,NULL,NULL,NULL,NULL},
	};
#endif

void lis_matvec_crs_u1_1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		#pragma omp parallel for private(i,j,is,ie,j0,t0)
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-0;j+=1)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-0;j+=1)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		#pragma omp parallel for private(i,j,is,ie,j0,t0)
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-0;j+=1)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvec_crs_u1_2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				t0 += A->L->value[j+ 0]*x[j0] + A->L->value[j+ 1]*x[j1];
			}
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				t0 += A->U->value[j+ 0]*x[j0] + A->U->value[j+ 1]*x[j1];
			}
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				t0 += vv0[j+ 0]*x[j0] + vv0[j+ 1]*x[j1];
			}
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvec_crs_u1_3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1,j2;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-2;j+=3)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				j2  = A->L->index[j+2];
				t0 += A->L->value[j+ 0]*x[j0] + A->L->value[j+ 1]*x[j1] + A->L->value[j+ 2]*x[j2];
			}
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-2;j+=3)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				j2  = A->U->index[j+2];
				t0 += A->U->value[j+ 0]*x[j0] + A->U->value[j+ 1]*x[j1] + A->U->value[j+ 2]*x[j2];
			}
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-2;j+=3)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				j2 = jj0[j+2];
				t0 += vv0[j+ 0]*x[j0] + vv0[j+ 1]*x[j1] + vv0[j+ 2]*x[j2];
			}
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvec_crs_u1_4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1,j2,j3;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-3;j+=4)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				j2  = A->L->index[j+2];
				j3  = A->L->index[j+3];
				t0 += A->L->value[j+ 0]*x[j0] + A->L->value[j+ 1]*x[j1] + A->L->value[j+ 2]*x[j2] + A->L->value[j+ 3]*x[j3];
			}
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-3;j+=4)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				j2  = A->U->index[j+2];
				j3  = A->U->index[j+3];
				t0 += A->U->value[j+ 0]*x[j0] + A->U->value[j+ 1]*x[j1] + A->U->value[j+ 2]*x[j2] + A->U->value[j+ 3]*x[j3];
			}
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-3;j+=4)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				j2 = jj0[j+2];
				j3 = jj0[j+3];
				t0 += vv0[j+ 0]*x[j0] + vv0[j+ 1]*x[j1] + vv0[j+ 2]*x[j2] + vv0[j+ 3]*x[j3];
			}
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvec_crs_u1_5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1,j2,j3,j4;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-4;j+=5)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				j2  = A->L->index[j+2];
				j3  = A->L->index[j+3];
				j4  = A->L->index[j+4];
				t0 += A->L->value[j+ 0]*x[j0] + A->L->value[j+ 1]*x[j1] + A->L->value[j+ 2]*x[j2] + A->L->value[j+ 3]*x[j3] + A->L->value[j+ 4]*x[j4];
			}
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-4;j+=5)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				j2  = A->U->index[j+2];
				j3  = A->U->index[j+3];
				j4  = A->U->index[j+4];
				t0 += A->U->value[j+ 0]*x[j0] + A->U->value[j+ 1]*x[j1] + A->U->value[j+ 2]*x[j2] + A->U->value[j+ 3]*x[j3] + A->U->value[j+ 4]*x[j4];
			}
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-4;j+=5)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				j2 = jj0[j+2];
				j3 = jj0[j+3];
				j4 = jj0[j+4];
				t0 += vv0[j+ 0]*x[j0] + vv0[j+ 1]*x[j1] + vv0[j+ 2]*x[j2] + vv0[j+ 3]*x[j3] + vv0[j+ 4]*x[j4];
			}
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvec_crs_u1_6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1,j2,j3,j4,j5;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-5;j+=6)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				j2  = A->L->index[j+2];
				j3  = A->L->index[j+3];
				j4  = A->L->index[j+4];
				j5  = A->L->index[j+5];
				t0 += A->L->value[j+ 0]*x[j0] + A->L->value[j+ 1]*x[j1] + A->L->value[j+ 2]*x[j2] + A->L->value[j+ 3]*x[j3] + A->L->value[j+ 4]*x[j4] + A->L->value[j+ 5]*x[j5];
			}
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-5;j+=6)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				j2  = A->U->index[j+2];
				j3  = A->U->index[j+3];
				j4  = A->U->index[j+4];
				j5  = A->U->index[j+5];
				t0 += A->U->value[j+ 0]*x[j0] + A->U->value[j+ 1]*x[j1] + A->U->value[j+ 2]*x[j2] + A->U->value[j+ 3]*x[j3] + A->U->value[j+ 4]*x[j4] + A->U->value[j+ 5]*x[j5];
			}
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-5;j+=6)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				j2 = jj0[j+2];
				j3 = jj0[j+3];
				j4 = jj0[j+4];
				j5 = jj0[j+5];
				t0 += vv0[j+ 0]*x[j0] + vv0[j+ 1]*x[j1] + vv0[j+ 2]*x[j2] + vv0[j+ 3]*x[j3] + vv0[j+ 4]*x[j4] + vv0[j+ 5]*x[j5];
			}
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvec_crs_u1_7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1,j2,j3,j4,j5,j6;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-6;j+=7)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				j2  = A->L->index[j+2];
				j3  = A->L->index[j+3];
				j4  = A->L->index[j+4];
				j5  = A->L->index[j+5];
				j6  = A->L->index[j+6];
				t0 += A->L->value[j+ 0]*x[j0] + A->L->value[j+ 1]*x[j1] + A->L->value[j+ 2]*x[j2] + A->L->value[j+ 3]*x[j3] + A->L->value[j+ 4]*x[j4] + A->L->value[j+ 5]*x[j5] + A->L->value[j+ 6]*x[j6];
			}
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-6;j+=7)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				j2  = A->U->index[j+2];
				j3  = A->U->index[j+3];
				j4  = A->U->index[j+4];
				j5  = A->U->index[j+5];
				j6  = A->U->index[j+6];
				t0 += A->U->value[j+ 0]*x[j0] + A->U->value[j+ 1]*x[j1] + A->U->value[j+ 2]*x[j2] + A->U->value[j+ 3]*x[j3] + A->U->value[j+ 4]*x[j4] + A->U->value[j+ 5]*x[j5] + A->U->value[j+ 6]*x[j6];
			}
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-6;j+=7)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				j2 = jj0[j+2];
				j3 = jj0[j+3];
				j4 = jj0[j+4];
				j5 = jj0[j+5];
				j6 = jj0[j+6];
				t0 += vv0[j+ 0]*x[j0] + vv0[j+ 1]*x[j1] + vv0[j+ 2]*x[j2] + vv0[j+ 3]*x[j3] + vv0[j+ 4]*x[j4] + vv0[j+ 5]*x[j5] + vv0[j+ 6]*x[j6];
			}
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvec_crs_u1_8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-7;j+=8)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				j2  = A->L->index[j+2];
				j3  = A->L->index[j+3];
				j4  = A->L->index[j+4];
				j5  = A->L->index[j+5];
				j6  = A->L->index[j+6];
				j7  = A->L->index[j+7];
				t0 += A->L->value[j+ 0]*x[j0] + A->L->value[j+ 1]*x[j1] + A->L->value[j+ 2]*x[j2] + A->L->value[j+ 3]*x[j3] + A->L->value[j+ 4]*x[j4] + A->L->value[j+ 5]*x[j5] + A->L->value[j+ 6]*x[j6] + A->L->value[j+ 7]*x[j7];
			}
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-7;j+=8)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				j2  = A->U->index[j+2];
				j3  = A->U->index[j+3];
				j4  = A->U->index[j+4];
				j5  = A->U->index[j+5];
				j6  = A->U->index[j+6];
				j7  = A->U->index[j+7];
				t0 += A->U->value[j+ 0]*x[j0] + A->U->value[j+ 1]*x[j1] + A->U->value[j+ 2]*x[j2] + A->U->value[j+ 3]*x[j3] + A->U->value[j+ 4]*x[j4] + A->U->value[j+ 5]*x[j5] + A->U->value[j+ 6]*x[j6] + A->U->value[j+ 7]*x[j7];
			}
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-7;j+=8)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				j2 = jj0[j+2];
				j3 = jj0[j+3];
				j4 = jj0[j+4];
				j5 = jj0[j+5];
				j6 = jj0[j+6];
				j7 = jj0[j+7];
				t0 += vv0[j+ 0]*x[j0] + vv0[j+ 1]*x[j1] + vv0[j+ 2]*x[j2] + vv0[j+ 3]*x[j3] + vv0[j+ 4]*x[j4] + vv0[j+ 5]*x[j5] + vv0[j+ 6]*x[j6] + vv0[j+ 7]*x[j7];
			}
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}




void lis_matvect_crs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n,np,nprocs;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	np   = A->np;
	if( A->is_splited )
	{
		for(i=0; i<np; i++)
		{
			y[i] = A->D->value[i] * x[i];
		}
		for(i=0; i<n; i++)
		{
			js = A->L->ptr[i];
			je = A->L->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->L->index[j];
				y[jj] += A->L->value[j] * t;
			}
		}
		for(i=0; i<n; i++)
		{
			js = A->U->ptr[i];
			je = A->U->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->U->index[j];
				y[jj] += A->U->value[j] * t;
			}
		}
	}
	else
	{
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<n; i++)
				{
					js = A->ptr[i];
					je = A->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
		#else
			for(i=0; i<np; i++)
			{
				y[i] = 0.0;
			}
			for(i=0; i<n; i++)
			{
				js = A->ptr[i];
				je = A->ptr[i+1];
				t = x[i];
				for(j=js;j<je;j++)
				{
					jj  = A->index[j];
					y[jj] += A->value[j] * t;
				}
			}
		#endif
	}
}

#ifdef USE_QUAD_PRECISION
void lis_matvec_crs_u1_1P(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
		int			i,j,k,h;
		int			n,np,nl,nz;
		int			is,ie,js,jj;
		int			is0;
		int			ie0;
		int			j0;
		int			*jj0;
		LIS_SCALAR			*vv0;
		LIS_SCALAR_M		tt0;
		LIS_MSCALAR		ax;
		LIS_SCALAR		*x,*y,*xl,*yl;
		LIS_QUAD_DECLAR;


		n     = A->n;
		x     = X->value;
		y     = Y->value;
		xl    = X->value_lo;
		yl    = Y->value_lo;
		tt0.hi = &X->work[0];
		tt0.lo = &X->work[1];
		if( A->is_splited )
		{
			for(i=0;i<n;i++)
			{
/*				lisqp_mul_dd_d(&t0,&x[i],A->D->value[i]);*/

				is = A->L->ptr[i];
				ie = A->L->ptr[i+1];
				for(j=is;j<ie-0;j+=1)
				{
					j0  = A->L->index[j+0];
/*					lisqp_mul_dd_d(&ax,&x[j0],A->L->value[j+ 0]);
					lisqp_add(&t0,&t0,&ax);*/
				}
				is = A->U->ptr[i];
				ie = A->U->ptr[i+1];
				for(j=is;j<ie-0;j+=1)
				{
					j0  = A->U->index[j+0];
/*					lisqp_mul_dd_d(&ax,&x[j0],A->U->value[j+ 0]);
					lisqp_add(&t0,&t0,&ax);*/
				}
/*				y[i] = t0;*/
			}
		}
		else
		{
			jj0 = A->index;
			vv0 = A->value;
			for(i=0;i<n;i++)
			{
				tt0.hi[0] = tt0.lo[0] = 0.0;

				is = A->ptr[i];
				ie = A->ptr[i+1];
				for(j=is;j<ie-0;j+=1)
				{
					j0 = jj0[j+0];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(tt0.hi[0],tt0.lo[0],tt0.hi[0],tt0.lo[0],x[j0],xl[j0],vv0[j]);
					#else
						LIS_QUAD_FMAD_SSE2_ALIGN(tt0.hi[0],tt0.lo[0],tt0.hi[0],tt0.lo[0],x[j0],xl[j0],vv0[j]);
					#endif
				}
				y[i]  = tt0.hi[0];
				yl[i] = tt0.lo[0];
			}
		}
	}

void lis_matvec_crs_u1_2P(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0,j1;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR_M	tt0;
	LIS_SCALAR_M	ax,xx,ax2,tt;
	LIS_SCALAR	*x,*y,*xl,*yl;
	LIS_QUAD_DECLAR;

	n     = A->n;
	x     = X->value;
	y     = Y->value;
	xl    = X->value_lo;
	yl    = Y->value_lo;
	tt.hi = &X->work[0];
	tt.lo = &X->work[2];
	xx.hi = &X->work[4];
	xx.lo = &X->work[6];
	ax2.hi = &X->work[8];
	ax2.lo = &X->work[10];
	ax.hi = &X->work[12];
	ax.lo = &X->work[13];
	tt0.hi = &X->work[14];
	tt0.lo = &X->work[15];
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
/*			lisqp_mul_dd_d(tt0.hi,&x[2*i],A->D->value[i]);*/

		/*	lisqp_zero(&tt,2);*/
			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0  = A->L->index[j+0];
				j1  = A->L->index[j+1];
				xx.hi[0] = x[j0];
				xx.lo[0] = xl[j0];
				xx.hi[1] = x[j1];
				xx.lo[1] = xl[j1];
/*				lisqp_mul_dd_d2_inline(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->L->value[j+ 0]);
				lisqp_add2_inline(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d2(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->L->value[j+ 0]);
				lisqp_add2(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d(&ax,&x[j0],A->L->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);
				lisqp_mul_dd_d(&ax,&x[j1],A->L->value[j+ 1]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
/*			lisqp_add(tt0.hi,tt.hi,tt.hi);*/
			for(;j<ie;j++)
			{
				j0  = A->L->index[j+0];
/*				lisqp_mul_dd_d(ax.hi,&x[2*j0],A->L->value[j+ 0]);
				lisqp_add(tt0.hi,tt0.hi,ax.hi);
				lisqp_mul_dd_d(&ax,&x[j0],A->L->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0  = A->U->index[j+0];
				j1  = A->U->index[j+1];
				xx.hi[0] = x[j0];
				xx.lo[0] = xl[j0];
				xx.hi[1] = x[j1];
				xx.lo[1] = xl[j1];
/*				lisqp_mul_dd_d2_inline(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->U->value[j+ 0]);
				lisqp_add2_inline(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d2(ax2.hi,ax2.lo,xx.hi,xx.lo,&A->U->value[j+ 0]);
				lisqp_add2(tt.hi,tt.lo,tt.hi,tt.lo,ax2.hi,ax2.lo);
				lisqp_mul_dd_d(&ax,&x[j0],A->U->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);
				lisqp_mul_dd_d(&ax,&x[j1],A->U->value[j+ 1]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
/*			lisqp_add(tt0.hi,tt.hi,tt.hi);*/
			for(;j<ie;j++)
			{
				j0  = A->U->index[j+0];
/*				lisqp_mul_dd_d(ax.hi,&x[2*j0],A->U->value[j+ 0]);
				lisqp_add(tt0.hi,tt0.hi,ax.hi);
				lisqp_mul_dd_d(&ax,&x[j0],A->U->value[j+ 0]);
				lisqp_add(&tt0,&tt0,&ax);*/
			}
			y[i]  = tt0.hi[0];
			yl[i] = tt0.lo[0];
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		for(i=0;i<n;i++)
		{
			tt0.hi[0] = tt0.lo[0] = 0.0;
			tt.hi[0] = tt.hi[1] = 0.0;
			tt.lo[0] = tt.lo[1] = 0.0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-1;j+=2)
			{
				j0 = jj0[j+0];
				j1 = jj0[j+1];
				#ifdef USE_SSE2
					LIS_QUAD_FMAD2_SSE2_ALIGN(tt.hi[0],tt.lo[0],tt.hi[0],tt.lo[0],x[j0],xl[j0],x[j1],xl[j1],vv0[j]);
				#endif
			}
			#ifdef USE_SSE2
				LIS_QUAD_ADD_SSE2(tt0.hi[0],tt0.lo[0],tt.hi[0],tt.lo[0],tt.hi[1],tt.lo[1]);
			#endif
			for(;j<ie;j++)
			{
				j0 = jj0[j+0];
				#ifdef USE_SSE2
					LIS_QUAD_FMAD_SSE2_ALIGN(tt0.hi[0],tt0.lo[0],tt0.hi[0],tt0.lo[0],x[j0],xl[j0],vv0[j]);
				#endif
			}
			y[i]  = tt0.hi[0];
			yl[i] = tt0.lo[0];
			
		}
	}
}

void lis_matvect_crs_mp(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	int i,j,k,js,je,jj;
	int n,np,nprocs;
	LIS_SCALAR_M tt0;
	LIS_SCALAR		*x,*y,*xl,*yl;
	LIS_SCALAR *w;
	LIS_QUAD_DECLAR;

	n    = A->n;
	np   = A->np;
	x     = X->value;
	y     = Y->value;
	xl    = X->value_lo;
	yl    = Y->value_lo;
	tt0.hi = &X->work[0];
	tt0.lo = &X->work[1];
	if( A->is_splited )
	{
		/*
		for(i=0; i<np; i++)
		{
			y[i] = A->D->value[i] * x[i];
		}
		for(i=0; i<n; i++)
		{
			js = A->L->ptr[i];
			je = A->L->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->L->index[j];
				y[jj] += A->L->value[j] * t;
			}
		}
		for(i=0; i<n; i++)
		{
			js = A->U->ptr[i];
			je = A->U->ptr[i+1];
			t = x[i];
			for(j=js;j<je;j++)
			{
				jj  = A->U->index[j];
				y[jj] += A->U->value[j] * t;
			}
		}
		*/
	}
	else
	{
		#ifdef _OPENMP
			/*
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<n; i++)
				{
					js = A->ptr[i];
					je = A->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
			*/
		#else
			for(i=0; i<np; i++)
			{
				y[i]  = 0.0;
				yl[i] = 0.0;
			}
			for(i=0; i<n; i++)
			{
				js = A->ptr[i];
				je = A->ptr[i+1];
				tt0.hi[0] = x[i];
				tt0.lo[0] = xl[i];
				for(j=js;j<je;j++)
				{
					jj  = A->index[j];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(y[jj],yl[jj],y[jj],yl[jj],tt0.hi[0],tt0.lo[0],A->value[j]);
					#else
						LIS_QUAD_FMAD_SSE2(y[jj],yl[jj],y[jj],yl[jj],tt0.hi[0],tt0.lo[0],A->value[j]);
					#endif
				}
			}
		#endif
	}
}

void lis_matvect_crs_mp2(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	int i,j,k,js,je,j0,j1;
	int n,np,nprocs;
	LIS_SCALAR_M tt0,yy0;
	LIS_SCALAR		*x,*y,*xl,*yl;
	LIS_SCALAR *w;
	LIS_QUAD_DECLAR;

	n    = A->n;
	np   = A->np;
	x     = X->value;
	y     = Y->value;
	xl    = X->value_lo;
	yl    = Y->value_lo;
	tt0.hi = &X->work[0];
	tt0.lo = &X->work[2];
	yy0.hi = &X->work[4];
	yy0.lo = &X->work[6];
	if( A->is_splited )
	{
	}
	else
	{
		#ifdef _OPENMP
		/*
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<n; i++)
				{
					js = A->ptr[i];
					je = A->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
			*/
		#else
			for(i=0; i<np; i++)
			{
				y[i]  = 0.0;
				yl[i] = 0.0;
			}
			for(i=0; i<n; i++)
			{
				js = A->ptr[i];
				je = A->ptr[i+1];
				tt0.hi[0] = x[i];
				tt0.hi[1] = x[i];
				tt0.lo[0] = xl[i];
				tt0.lo[1] = xl[i];
				for(j=js;j<je-1;j+=2)
				{
					j0  = A->index[j];
					j1  = A->index[j+1];
					#ifdef USE_SSE2
						LIS_QUAD_FMAD2_SSE2_ALIGN2(y[j0],yl[j0],y[j1],yl[j1],tt0.hi[0],tt0.lo[0],A->value[j]);
					#endif
				}
				for(;j<je;j++)
				{
					j0  = A->index[j];
					#ifdef USE_SSE2
						LIS_QUAD_FMAD_SSE2(y[j0],yl[j0],y[j0],yl[j0],tt0.hi[0],tt0.lo[0],A->value[j]);
					#endif
				}
			}
		#endif
	}
}
#endif

