/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

void lis_matvec_dia(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,ii;
	int n,np,nnd,k;
	int my_rank,nprocs;
	LIS_SCALAR t;

	if( A->is_splited )
	{
		n      = A->n;
		np     = A->np;
		#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
		{
			#ifdef _OPENMP
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
			#else
				nprocs  = 1;
				my_rank = 0;
			#endif
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie)

			#pragma cdir nodep
			for(i=is;i<ie;i++)
			{ 
				y[i] = A->D->value[i] * x[i];
			}
			for(j=0;j<A->L->nnd;j++)
			{
				jj = A->L->index[j];
				js = _max(is,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(ie,np-jj);
				#else
					je = _min(ie,n-jj);
				#endif
				k  = is*A->L->nnd + j*(ie-is);
				ii = js-is;
				#pragma cdir nodep
				for(i=js;i<je;i++)
				{ 
					y[i] += A->L->value[k + ii] * x[jj+i];
					ii++;
				}
			}
			for(j=0;j<A->U->nnd;j++)
			{
				jj = A->U->index[j];
				js = _max(is,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(ie,np-jj);
				#else
					je = _min(ie,n-jj);
				#endif
				k  = is*A->U->nnd + j*(ie-is);
				ii = js-is;
				#pragma cdir nodep
				for(i=js;i<je;i++)
				{ 
					y[i] += A->U->value[k + ii] * x[jj+i];
					ii++;
				}
			}
		}
	}
	else
	{
		n      = A->n;
		np     = A->np;
		nnd    = A->nnd;

		#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
		{
			#ifdef _OPENMP
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
			#else
				nprocs  = 1;
				my_rank = 0;
			#endif
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie)

			#pragma cdir nodep
			for(i=is; i<ie; i++)
			{
				y[i] = 0.0;
			}
			for(j=0;j<nnd;j++)
			{
				jj = A->index[j];
				js = _max(is,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(ie,np-jj);
				#else
					je = _min(ie,n-jj);
				#endif
				k  = is*nnd + j*(ie-is);
				ii = js-is;
				#pragma cdir nodep
				for(i=js;i<je;i++)
				{ 
					y[i] += A->value[k + ii] * x[jj+i];
					ii++;
				}
			}
		}
	}
}

void lis_matvect_dia(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,ii;
	int n,nnd,np,k;
	int my_rank,nprocs;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	if( A->is_splited )
	{
		n      = A->n;
		#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
		{
			#ifdef _OPENMP
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
			#else
				nprocs  = 1;
				my_rank = 0;
			#endif
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie)

			#pragma cdir nodep
			for(i=is; i<ie; i++)
			{
				y[i] = 0.0;
			}
			for(j=0;j<A->L->nnd;j++)
			{
				jj = A->L->index[j];
				js = _max(is,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(ie,np-jj);
				#else
					je = _min(ie,n-jj);
				#endif
				k  = is*A->L->nnd + j*(ie-is);
				ii = js-is;
				#pragma cdir nodep
				for(i=js;i<je;i++)
				{ 
					y[jj+i] += A->L->value[k + ii] * x[i];
					ii++;
				}
			}
			#pragma cdir nodep
			for(i=is;i<ie;i++)
			{ 
				y[i] += A->D->value[i] * x[i];
			}
			for(j=0;j<A->U->nnd;j++)
			{
				jj = A->U->index[j];
				js = _max(is,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?ie:_min(ie,np-jj);
				#else
					je = _min(ie,n-jj);
				#endif
				k  = is*A->U->nnd + j*(ie-is);
				ii = js-is;
				#pragma cdir nodep
				for(i=js;i<je;i++)
				{ 
					y[jj+i] += A->U->value[k + ii] * x[i];
					ii++;
				}
			}
		}
	}
	else
	{
		n      = A->n;
		np     = A->np;
		nnd    = A->nnd;

		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
			{
				nprocs  = omp_get_max_threads();
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie)

				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				for(j=0;j<nnd;j++)
				{
					jj = A->index[j];
					js = _max(is,-jj);
					#ifdef USE_MPI
						je = jj<=(np-n)?ie:_min(ie,np-jj);
					#else
						je = _min(ie,n-jj);
					#endif
					k  = is*nnd + j*(ie-is);
					ii = js-is;
					#pragma cdir nodep
					for(i=js;i<je;i++)
					{ 
						w[my_rank*np + jj+i] += A->value[k + ii] * x[i];
						ii++;
					}
				}
				#pragma omp barrier
				#pragma cdir nodep
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
		#else
			#pragma cdir nodep
			for(i=0;i<np;i++)
			{
				y[i] = 0.0;
			}
			for(j=0;j<nnd;j++)
			{
				jj = A->index[j];
				js = _max(0,-jj);
				#ifdef USE_MPI
					je = jj<=(np-n)?n:_min(n,np-jj);
				#else
					je = _min(n,n-jj);
				#endif
				k  = j*n;
				ii = js-is;
				#pragma cdir nodep
				for(i=js;i<je;i++)
				{ 
					y[jj+i] += A->value[k + ii] * x[i];
					ii++;
				}
			}
		#endif
	}
}
