/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

#undef __FUNC__
#define __FUNC__ "lis_matvec_vbr"
void lis_matvec_vbr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int bi,bj,bc;
	int nr,nc;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	if( A->is_splited )
	{
		LIS_SETERR_IMP;
		return;
	}
	else
	{
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			y[i] = 0.0;
		}
		#pragma omp parallel for private(i,j,k,bi,bj,bc)
		for(bi=0;bi<nr;bi++)
		{
			for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
			{
				bj   = A->bindex[bc];
				k    = A->ptr[bc];
				for(j=A->col[bj];j<A->col[bj+1];j++)
				{
					for(i=A->row[bi];i<A->row[bi+1];i++)
					{
						y[i] += A->value[k] * x[j];
						k++;
					}
				}
			}
		}
	}
}


#undef __FUNC__
#define __FUNC__ "lis_matvect_vbr"
void lis_matvect_vbr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n,np,nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n   = A->n;
	np  = A->np;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;

	if( A->is_splited )
	{
		LIS_SETERR_IMP;
		return;
	}
	else
	{
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR) );
			#pragma omp parallel private(bi,bc,bj,i,j,k,my_rank)
			{
				my_rank = omp_get_thread_num();

				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(bi=0;bi<nr;bi++)
				{
					for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
					{
						bj   = A->bindex[bc];
						k    = A->ptr[bc];
						for(j=A->col[bj];j<A->col[bj+1];j++)
						{
							for(i=A->row[bi];i<A->row[bi+1];i++)
							{
								w[my_rank*np + j] += A->value[k] * x[i];
								k++;
							}
						}
					}
				}
				#pragma omp barrier
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
		#else
			for(i=0; i<n; i++)
			{
				y[i] = 0.0;
			}
			for(bi=0;bi<nr;bi++)
			{
				for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
				{
					bj   = A->bindex[bc];
					k    = A->ptr[bc];
					for(j=A->col[bj];j<A->col[bj+1];j++)
					{
						for(i=A->row[bi];i<A->row[bi+1];i++)
						{
							y[j] += A->value[k] * x[i];
							k++;
						}
					}
				}
			}
		#endif
	}
}
