/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/

#if defined(USE_SAAMG)

extern void F77_FUNC_(finit_data_creation,FINIT_DATA_CREATION)(void *c_data_creation_ptr_bar);
extern void F77_FUNC_(finit_data_creation_unsym,FINIT_DATA_CREATION_UNSYM)(void *c_data_creation_unsym_ptr_bar);
extern void F77_FUNC_(finit_v_cycle,FINIT_V_CYCLE)(void *c_v_cycle_ptr_bar);
extern void F77_FUNC_(finit_clear_matrix,FINIT_CLEAR_MATRIX)(void *c_clear_matrix_ptr_bar);

char *f_data_creation_ptr;
char *f_data_creation_unsym_ptr;
char *f_v_cycle_ptr;
char *f_clear_matrix_ptr;
void c_data_creation_ptr_bar(char *bar)
{
	f_data_creation_ptr = bar;
}
void c_data_creation_unsym_ptr_bar(char *bar)
{
	f_data_creation_unsym_ptr = bar;
}
void c_v_cycle_ptr_bar(char *bar)
{
	f_v_cycle_ptr = bar;
}
void c_clear_matrix_ptr_bar(char *bar)
{
	f_clear_matrix_ptr = bar;
}
#endif

#undef __FUNC__
#define __FUNC__ "lis_precon_create_saamg"
int lis_precon_create_saamg(LIS_SOLVER solver, LIS_PRECON precon)
{
#if defined(USE_SAAMG)
	LIS_MATRIX	A;
	LIS_COMMTABLE	table;
	int			i;
	int			unsym,sol;
	int			err;
	#ifdef USE_MPI
		MPI_Fint	comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	if( solver->A->matrix_type!=LIS_MATRIX_CRS )
	{
		LIS_SETERR_IMP;
		return LIS_ERR_NOT_IMPLEMENTED;
	}
    precon->A       = solver->A;
    precon->is_copy = LIS_FALSE;
    A               = precon->A;
	sol             = solver->options[LIS_OPTIONS_SOLVER];
	unsym           = solver->options[LIS_OPTIONS_SAAMG_UNSYM];
#if 0
	if( sol!=LIS_SOLVER_CG && !unsym )
	{
		unsym = LIS_TRUE;
	}
#endif

	err = lis_vector_duplicate(A,&precon->temp);
	if( err )
	{
		return err;
	}
	F77_FUNC_(finit_data_creation,FINIT_DATA_CREATION)(c_data_creation_ptr_bar);
	F77_FUNC_(finit_data_creation_unsym,FINIT_DATA_CREATION_UNSYM)(c_data_creation_unsym_ptr_bar);
	F77_FUNC_(finit_v_cycle,FINIT_V_CYCLE)(c_v_cycle_ptr_bar);
	F77_FUNC_(finit_clear_matrix,FINIT_CLEAR_MATRIX)(c_clear_matrix_ptr_bar);
	
	lis_matrix_split(A);

	#ifdef USE_MPI
		comm = MPI_Comm_c2f(A->comm);
		lis_send_recv(A->commtable,A->D->value);
		table = A->commtable;
		if( !unsym )
		{
			(*(void (*)())f_data_creation_ptr)(&A->n,&A->np,&A->L->nnz,&A->U->nnz,
				A->D->value,A->L->value,A->L->ptr,A->L->index,
				A->U->value, A->U->ptr, A->U->index,
				&table->neibpetot, table->neibpe, table->import_ptr,
				table->import_index, table->export_ptr, table->export_index,
				&table->imnnz,&table->exnnz,
				&comm, &precon->level_num,&precon->wsize);
		}
		else
		{
			(*(void (*)())f_data_creation_unsym_ptr)(&A->n,&A->np,&A->L->nnz,&A->U->nnz,
				A->D->value,A->L->value,A->L->ptr,A->L->index,
				A->U->value, A->U->ptr, A->U->index,
				&table->neibpetot, table->neibpe, table->import_ptr,
				table->import_index, table->export_ptr, table->export_index,
				&table->imnnz,&table->exnnz,
				&comm, &precon->level_num,&precon->wsize);
		}
	#else
		if( !unsym )
		{
			(*(void (*)())f_data_creation_ptr)(&A->n,&A->L->nnz,&A->U->nnz,
				A->D->value,A->L->value,A->L->ptr,A->L->index,
				A->U->value, A->U->ptr, A->U->index, &precon->level_num);
		}
		else
		{
			(*(void (*)())f_data_creation_unsym_ptr)(&A->n,&A->L->nnz,&A->U->nnz,
				A->D->value,A->L->value,A->L->ptr,A->L->index,
				A->U->value, A->U->ptr, A->U->index, &precon->level_num);
		}
	#endif

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
#else
	LIS_DEBUG_FUNC_IN;

    precon->A       = solver->A;
    precon->is_copy = LIS_FALSE;

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
#endif
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_saamg"
int lis_psolve_saamg(LIS_SOLVER solver, LIS_VECTOR b, LIS_VECTOR x)
{
#if defined(USE_SAAMG)
	int		n;
	LIS_PRECON precon;
	LIS_MATRIX A;
	#ifdef USE_MPI
		MPI_Fint comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	A      = solver->A;

	#ifdef USE_MPI
		comm = MPI_Comm_c2f(A->comm);
		n = b->np;
		 (*(void (*)())f_v_cycle_ptr)(b->value,x->value,precon->temp->value,&precon->level_num,
			 &comm,A->commtable->ws,A->commtable->wr,&n,&precon->wsize);
	#else
		n = b->n;
		 (*(void (*)())f_v_cycle_ptr)(&n,b->value,x->value,&precon->level_num,precon->temp->value);
	#endif
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	LIS_DEBUG_FUNC_IN;

	lis_vector_copy(b,x);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_saamg"
int lis_psolvet_saamg(LIS_SOLVER solver, LIS_VECTOR b, LIS_VECTOR x)
{
#if defined(USE_SAAMG)
	int		n;
	LIS_PRECON precon;
	LIS_MATRIX A;
	#ifdef USE_MPI
		MPI_Fint comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	A      = solver->A;
	precon = solver->precon;

	#ifdef USE_MPI
		comm = MPI_Comm_c2f(A->comm);
		n = b->np;
		 (*(void (*)())f_v_cycle_ptr)(b->value,x->value,precon->temp->value,&precon->level_num,
			 &comm,A->commtable->ws,A->commtable->wr,&n,&precon->wsize);
	#else
		n = b->n;
		 (*(void (*)())f_v_cycle_ptr)(&n,b->value,x->value,&precon->level_num,precon->temp->value);
	#endif
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	LIS_DEBUG_FUNC_IN;

	lis_vector_copy(b,x);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}
