/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

char *LIS_ERR_MESS[] = {
	"ILL_ARG",
	"OUT_OF_MEMORY",
	"NOT_IMPLEMENTED",
	"FILE_IO_ERROR"
};

LIS_Comm lis_debug_comm = LIS_COMM_WORLD;

void lis_debug_set_comm(LIS_Comm comm)
{
	lis_debug_comm = LIS_COMM_WORLD;
}

int lis_debug_trace_func(int flag, char *func)
{
	static int lis_tr_func = 0;
	char buf[1024];
	int argc=0;

	#ifdef USE_MPI
		MPI_Initialized(&lis_mpi_initialized);
		if (!lis_mpi_initialized) MPI_Init(&argc, NULL);
	#endif
	if( flag )
	{
		sprintf(buf, "%%%ds : %%s\n",lis_tr_func+3);
		lis_printf(lis_debug_comm, buf, "IN ", func);
		lis_tr_func++;
	}
	else
	{
		lis_tr_func--;
		sprintf(buf, "%%%ds : %%s\n",lis_tr_func+3);
		lis_printf(lis_debug_comm, buf, "OUT", func);
	}
	return LIS_SUCCESS;
}

int lis_printf(LIS_Comm comm, const char *mess, ...)
{
	va_list vvlist;
	int		my_rank;

	#ifdef USE_MPI
	    MPI_Comm_rank(comm,&my_rank);
	#else
		my_rank = 0;
	#endif

	if( my_rank==0 )
	{
		va_start( vvlist, mess );
		vprintf(mess,vvlist);
		va_end(vvlist);
	}
	return LIS_SUCCESS;
}


int lis_error(const char *file, const char *func, const int line, const int code, const char *mess, ...)
{
	va_list vvlist;

	va_start( vvlist, mess );
	lis_printf(lis_debug_comm,"%s(%d) : %s : error %s :",file,line,func,LIS_ERR_MESS[code-LIS_ERR_ILL_ARG]);
	if( mess ) vprintf(mess,vvlist);
	va_end(vvlist);
	return LIS_SUCCESS;
}

void CHKERR(int err)
{
	if(err)
	{
		lis_finalize();
		exit(err);
	}
}

