/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif
#undef malloc

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include "lislib.h"

typedef struct _malloc_address {
  struct _malloc_address *next, *prev;
  void *address, *real_address;
  size_t size;
} malloc_address;

malloc_address malloc_address_top = 
  {&malloc_address_top, &malloc_address_top, NULL, NULL};

void *malloc ();
 
void *rpl_malloc( size_t n )
{
	if( n == 0 )
	n = 1;
	return malloc( n );
}

void *lis_malloc( size_t size )
{
  union {
    void *ptr;
    unsigned long l;
  } addr;
  malloc_address *ma;

	addr.ptr = malloc(size + 16);
    ma = malloc(sizeof(malloc_address));
    ma->next = &malloc_address_top;
	ma->prev = malloc_address_top.prev;
	ma->prev->next = ma;
	ma->next->prev = ma;
	ma->real_address = addr.ptr;
	ma->size = size;
	addr.l += 0xf;
	addr.l &= ~0xf;
	ma->address = addr.ptr;
 
  return addr.ptr;
}

void *lis_calloc( size_t size )
{
  union {
    void *ptr;
    unsigned long l;
  } addr;
  malloc_address *ma;

	addr.ptr = malloc(size + 16);
    ma = malloc(sizeof(malloc_address));
    ma->next = &malloc_address_top;
	ma->prev = malloc_address_top.prev;
	ma->prev->next = ma;
	ma->next->prev = ma;
	ma->real_address = addr.ptr;
	ma->size = size;
	addr.l += 0xf;
	addr.l &= ~0xf;
	ma->address = addr.ptr;
	memset(addr.ptr,0,size);
 
  return addr.ptr;
}

void *lis_realloc( void *p, size_t size )
{
	malloc_address *ma;
	size_t old_size;
	void *s;

	ma = malloc_address_top.next;
	while (ma->address)
	{
	    if (ma->address == p) break;
		ma = ma->next;
	}
	if (ma->address)
	{
		old_size = ma->size;
		ma->next->prev = ma->prev;
		ma->prev->next = ma->next;
		s = lis_malloc(size);
		memcpy(s,p,old_size);
		free(ma->real_address);
		free(ma);
		return s;
	}
	s = realloc(p,size);
	return s;
}

void lis_free(void *p)
{
  malloc_address *ma;

  ma = malloc_address_top.next;
  while (ma->address) {
    if (ma->address == p) break;
	ma = ma->next;
  }
  if (ma->address) {
    ma->next->prev = ma->prev;
	ma->prev->next = ma->next;
    free(ma->real_address);
	free(ma);
	return;
  }
  free(p);
  return;
}

void lis_free2(int n, ...)
{
	va_list vvlist;
	int		i;
	char	*p;

	va_start( vvlist, n );
	for(i=0;i<n;i++)
	{
		p = va_arg(vvlist, char *);
		if( p )
		{
			lis_free(p);
		}
	}
	va_end(vvlist);
}

int lis_is_malloc( void *p )
{
	malloc_address *ma;

	ma = malloc_address_top.next;
	while (ma->address)
	{
	    if (ma->address == p) break;
		ma = ma->next;
	}
	if (ma->address)
	{
		return LIS_TRUE;
	}
	return LIS_FALSE;
}
