/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_output
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_output"
int lis_output(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				err;
	FILE			*file;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_ALL);
	if( err ) return err;

	if( format==LIS_FMT_MM || format==LIS_FMT_MMB )
	{
		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			err = lis_output_mm_crs(A,b,x,format,path);
			break;
/*
		case LIS_MATRIX_CCS:
			err = lis_output_mm_ccs(A,b,x,path);
			break;
		case LIS_MATRIX_MSR:
			err = lis_output_mm_msr(A,b,x,path);
			break;
		case LIS_MATRIX_DIA:
			err = lis_output_mm_dia(A,b,x,path);
			break;
		case LIS_MATRIX_ELL:
			err = lis_output_mm_ell(A,b,x,path);
			break;
		case LIS_MATRIX_JDS:
			err = lis_output_mm_jds(A,b,x,path);
			break;
		case LIS_MATRIX_BSR:
			err = lis_output_mm_bsr(A,b,x,path);
			break;
		case LIS_MATRIX_BSC:
			err = lis_output_mm_bsc(A,b,x,path);
			break;
		case LIS_MATRIX_VBR:
			err = lis_output_mm_vbr(A,b,x,path);
			break;
		case LIS_MATRIX_DNS:
			err = lis_output_mm_dns(A,b,x,path);
			break;
		case LIS_MATRIX_COO:
			err = lis_output_mm_coo(A,b,x,path);
			break;
*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}
	else
	{
		switch( A->matrix_type )
		{
/*
		case LIS_MATRIX_CRS:
			err = lis_output_lis_crs(A,b,x,format,path);
			break;
		case LIS_MATRIX_CCS:
			err = lis_output_lis_ccs(A,b,x,format,path);
			break;
		case LIS_MATRIX_MSR:
			err = lis_output_lis_msr(A,b,x,format,path);
			break;
		case LIS_MATRIX_DIA:
			err = lis_output_lis_dia(A,b,x,format,path);
			break;
		case LIS_MATRIX_ELL:
			err = lis_output_lis_ell(A,b,x,format,path);
			break;
		case LIS_MATRIX_JDS:
			err = lis_output_lis_jds(A,b,x,format,path);
			break;
		case LIS_MATRIX_BSR:
			err = lis_output_lis_bsr(A,b,x,format,path);
			break;
		case LIS_MATRIX_BSC:
			err = lis_output_lis_bsc(A,b,x,format,path);
			break;
		case LIS_MATRIX_VBR:
			err = lis_output_lis_vbr(A,b,x,format,path);
			break;
		case LIS_MATRIX_DNS:
			err = lis_output_lis_dns(A,b,x,format,path);
			break;
		case LIS_MATRIX_COO:
			err = lis_output_lis_coo(A,b,x,format,path);
			break;
*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return err;
}

#undef __FUNC__
#define __FUNC__ "lis_output_vector"
int lis_output_vector(LIS_VECTOR v, int format, char *filename)
{
	int				err;
	FILE			*file;

	LIS_DEBUG_FUNC_IN;

	err = lis_vector_check(v,LIS_VECTOR_CHECK_NULL);
	if( err ) return err;

	switch( format )
	{
	case LIS_FMT_PLAIN:
		err = lis_output_vector_plain(v,filename);
		break;
	case LIS_FMT_MM:
		err = lis_output_vector_mm(v,filename);
		break;
	case LIS_FMT_LIS:
		err = lis_output_vector_lis_ascii(v,filename);
		break;
	default:
		LIS_SETERR(LIS_ERR_ILL_ARG,"ill format option\n");
		return LIS_ERR_ILL_ARG;
		break;
	}

	LIS_DEBUG_FUNC_OUT;
	return err;
}

#undef __FUNC__
#define __FUNC__ "lis_output_vector_plain"
int lis_output_vector_plain(LIS_VECTOR v, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err,ret;
	FILE			*file;

	LIS_DEBUG_FUNC_IN;

#ifdef USE_MPI
	nprocs    = v->nprocs;
	my_rank   = v->my_rank;
	n         = v->n;

	err = 0;
	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			if( my_rank==0 )
			{
				file = fopen(path, "w");
			}
			else
			{
				file = fopen(path, "a");
			}
			if( file==NULL )
			{
				LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
				err = 1;
			}
			for(i=0;i<n;i++)
			{
				fprintf(file, "%28.20e\n", v->value[i]);
			}
			fclose(file);
		}
		MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
		if( ret )
		{
			return LIS_FAILS;
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	n         = v->n;

	file = fopen(path, "w");
	if( file==NULL )
	{
		LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
		return LIS_ERR_FILE_IO;
	}
	for(i=0;i<n;i++)
	{
		fprintf(file, "%28.20e\n", v->value[i]);
	}
	fclose(file);
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}

#undef __FUNC__
#define __FUNC__ "lis_output_vector_mm"
int lis_output_vector_mm(LIS_VECTOR v, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe,is;
	int				err,ret;
	FILE			*file;

	LIS_DEBUG_FUNC_IN;
#ifdef USE_MPI
	nprocs    = v->nprocs;
	my_rank   = v->my_rank;
	n         = v->n;
	is        = v->is;

	err = 0;
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
			err = 1;
		}

		fprintf(file, "%%%%MatrixMarket vector coordinate real general\n");
		fprintf(file, "%d\n", v->gn);
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	err = 0;
	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
				err = 1;
			}
			else
			{
				for(i=0;i<n;i++)
				{
					fprintf(file, "%d %28.20e\n", i+is+1, v->value[i]);
				}
				fclose(file);
			}
		}
		MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
		if( ret )
		{
			return LIS_FAILS;
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	n         = v->n;
	is        = v->is;

	file = fopen(path, "w");
	if( file==NULL )
	{
		LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
		return LIS_ERR_FILE_IO;
	}

	fprintf(file, "%%%%MatrixMarket vector coordinate real general\n");
	fprintf(file, "%d\n", v->gn);

	for(i=0;i<n;i++)
	{
		fprintf(file, "%d %28.20e\n", i+is+1, v->value[i]);
	}
	fclose(file);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}


#undef __FUNC__
#define __FUNC__ "lis_output_vector_lis_ascii"
int lis_output_vector_lis_ascii(LIS_VECTOR v, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err,ret;
	FILE			*file;

	LIS_DEBUG_FUNC_IN;
#ifdef USE_MPI
	nprocs    = v->nprocs;
	my_rank   = v->my_rank;
	n         = v->n;

	err = 0;
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
			err = 1;
		}

		fprintf(file, "#LIS A vec\n");
		fprintf(file, "%d\n", nprocs);
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	err = 0;
	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
				err = 1;
			}
			else
			{
				fprintf(file, "# %d %d\n", pe, v->n);
				for(i=0;i<n;i++)
				{
					fprintf(file, "%28.20e ",v->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( n%3!=0 ) fprintf(file, "\n");
				fclose(file);
			}
		}
		MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
		if( ret )
		{
			return LIS_FAILS;
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	n         = v->n;

	file = fopen(path, "w");
	if( file==NULL )
	{
		LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
		return LIS_ERR_FILE_IO;
	}

	fprintf(file, "#LIS A vec\n");
	fprintf(file, "1\n");

	fprintf(file, "# 0 %d\n", v->n);
	for(i=0;i<n;i++)
	{
		fprintf(file, "%28.20e ",v->value[i]);
		if( (i+1)%3==0 ) fprintf(file, "\n");
	}
	if( n%3!=0 ) fprintf(file, "\n");
	fclose(file);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}

#undef __FUNC__
#define __FUNC__ "lis_output_residual_history"
int lis_output_residual_history(LIS_SOLVER solver, char *filename)
{
	int		i,my_rank,err;
	FILE	*file;

	LIS_DEBUG_FUNC_IN;
#ifdef USE_MPI
	if( solver->residual==NULL )
	{
		LIS_SETERR(LIS_FAILS,"residual history is empty\n");
		return LIS_FAILS;
	}
	if( solver->A==NULL )
	{
		LIS_SETERR(LIS_FAILS,"matrix A is NULL\n");
		return LIS_FAILS;
	}

	my_rank = solver->A->my_rank;
	err = 0;
	if( my_rank==0 )
	{
		file = fopen(filename, "w");
		if( file==NULL )
		{
			LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", filename);
			err = 1;
		}
		else
		{
			for(i=0;i<solver->iter;i++)
			{
				fprintf(file, "%e\n", solver->residual[i]);
			}
			fclose(file);
		}
	}
	MPI_Barrier(solver->A->comm);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	if( solver->residual==NULL )
	{
		LIS_SETERR(LIS_FAILS,"residual history is empty\n");
		return LIS_FAILS;
	}
	file = fopen(filename, "w");
	if( file==NULL )
	{
		LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", filename);
		return LIS_ERR_FILE_IO;
	}
	for(i=0;i<solver->iter;i++)
	{
		fprintf(file, "%e\n", solver->residual[i]);
	}
	fclose(file);
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}
