/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#ifdef WIN32
	#include <windows.h>
#else
	#include <sys/time.h>
#endif

#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif


#ifdef USE_MPI
double lis_wtime(void)
{
	double  etime;
	
	etime = MPI_Wtime();

	return etime;
}
#else
#ifdef _OPENMP
double lis_wtime(void)
{
	double  etime;
	
	etime = omp_get_wtime();

	return etime;
}
#else
#ifdef WIN32
static LARGE_INTEGER t = {0, 0};
double lis_wtime(void)
{

  LARGE_INTEGER l;
  if (t.QuadPart == 0)
    QueryPerformanceFrequency(&t);
  QueryPerformanceCounter(&l);
  return (double)l.QuadPart / (double)t.QuadPart;
}
#else
#ifdef HAVE_CLOCK_GETTIME
double lis_wtime(void)
{
	double  etime;
	struct timespec ts;

	clock_gettime(CLOCK_REALTIME, &ts);
	etime = tv.tv_sec + (double)tv.tv_usec*1.0e-9;
	return etime;
}
#else
double lis_wtime(void)
{
	double  etime;
	struct timeval tv;

	gettimeofday(&tv,NULL);
	etime = tv.tv_sec + (double)tv.tv_usec*1.0e-6;
	return etime;
}
#endif
#endif
#endif
#endif
